Package: ao
Title: Alternating Optimization
Version: 1.2.3
Authors@R: 
    c(
    person(given = "Lennart",
           family = "Oelschl\U00E4ger",
           role = c("aut", "cre"),
           email = "oelschlaeger.lennart@gmail.com",
           comment = c(ORCID = "0000-0001-5421-9313")),
    person(given = "Siddhartha",
           family = "Chib",
           role = "ctb",
           email = "chib@wustl.edu")
    )
Description: Implementation of an iterative process that optimizes a function by
    alternately performing restricted optimization over parameter subsets.
    Instead of solving one joint optimization problem, alternating optimization
    breaks it into smaller sub-problems. This approach can make optimization
    feasible when joint optimization is too difficult.
URL: https://loelschlaeger.de/ao/, https://github.com/loelschlaeger/ao/
BugReports: https://github.com/loelschlaeger/ao/issues
License: GPL-3
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Imports: checkmate, cli, future.apply, oeli (>= 0.7.5), optimizeR (>=
        1.2.1), progressr, R6, stats, utils
Suggests: devtools, ggplot2, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
Depends: R (>= 4.1.0)
NeedsCompilation: no
Packaged: 2026-05-11 14:44:03 UTC; Lennart Oelschläger
Author: Lennart Oelschläger [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5421-9313>),
  Siddhartha Chib [ctb]
Maintainer: Lennart Oelschläger <oelschlaeger.lennart@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-11 17:10:09 UTC
Built: R 4.6.0; ; 2026-05-11 20:37:03 UTC; unix

Package: ao
Title: Alternating Optimization
Version: 0.3.3
Authors@R: 
    person(given = "Lennart",
           family = "Oelschl\U00E4ger",
           role = c("aut", "cre"),
           email = "oelschlaeger.lennart@gmail.com",
           comment = c(ORCID = "0000-0001-5421-9313"))
Description: Alternating optimization of (high-dimensional) functions is an 
    iterative procedure for optimizing jointly over all parameters by 
    alternately optimizing parameter subsets.
URL: https://loelschlaeger.de/ao/, https://github.com/loelschlaeger/ao/
BugReports: https://github.com/loelschlaeger/ao/issues
License: GPL-3
Encoding: UTF-8
RoxygenNote: 7.3.1
Imports: checkmate, cli, oeli, utils
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
Depends: R (>= 4.0.0), optimizeR (>= 1.0.4)
NeedsCompilation: no
Packaged: 2024-02-29 20:05:27 UTC; Oelschlaeger
Author: Lennart Oelschläger [aut, cre]
    (<https://orcid.org/0000-0001-5421-9313>)
Maintainer: Lennart Oelschläger <oelschlaeger.lennart@gmail.com>
Repository: CRAN
Date/Publication: 2024-02-29 21:12:46 UTC
Built: R 4.2.3; ; 2024-03-04 12:19:35 UTC; unix

Package: ao
Title: Alternating Optimization
Version: 1.2.3
Authors@R: 
    c(
    person(given = "Lennart",
           family = "Oelschl\U00E4ger",
           role = c("aut", "cre"),
           email = "oelschlaeger.lennart@gmail.com",
           comment = c(ORCID = "0000-0001-5421-9313")),
    person(given = "Siddhartha",
           family = "Chib",
           role = "ctb",
           email = "chib@wustl.edu")
    )
Description: Implementation of an iterative process that optimizes a function by
    alternately performing restricted optimization over parameter subsets.
    Instead of solving one joint optimization problem, alternating optimization
    breaks it into smaller sub-problems. This approach can make optimization
    feasible when joint optimization is too difficult.
URL: https://loelschlaeger.de/ao/, https://github.com/loelschlaeger/ao/
BugReports: https://github.com/loelschlaeger/ao/issues
License: GPL-3
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Imports: checkmate, cli, future.apply, oeli (>= 0.7.5), optimizeR (>=
        1.2.1), progressr, R6, stats, utils
Suggests: devtools, ggplot2, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
Depends: R (>= 4.1.0)
NeedsCompilation: no
Packaged: 2026-05-11 14:44:03 UTC; Lennart Oelschläger
Author: Lennart Oelschläger [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5421-9313>),
  Siddhartha Chib [ctb]
Maintainer: Lennart Oelschläger <oelschlaeger.lennart@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-11 17:10:09 UTC

# Generated by roxygen2: do not edit by hand

export(ao)
importFrom(R6,R6Class)
importFrom(checkmate,assert_int)
importFrom(checkmate,assert_number)
importFrom(checkmate,test_class)
importFrom(checkmate,test_flag)
importFrom(checkmate,test_function)
importFrom(checkmate,test_list)
importFrom(checkmate,test_number)
importFrom(cli,cat_line)
importFrom(cli,cli_abort)
importFrom(cli,style_hyperlink)
importFrom(future.apply,future_lapply)
importFrom(oeli,match_arg)
importFrom(oeli,test_numeric_vector)
importFrom(optimizeR,Objective)
importFrom(optimizeR,Optimizer)
importFrom(progressr,progressor)
importFrom(stats,rnorm)
importFrom(utils,packageVersion)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.path = "ao-"
)
library("ao")
set.seed(2)

## ----ao call, eval = FALSE----------------------------------------------------
# ao(
#   f,
#   initial,
#   target = NULL,
#   npar = NULL,
#   gradient = NULL,
#   hessian = NULL,
#   ...,
#   partition = "sequential",
#   new_block_probability = 0.3,
#   minimum_block_number = 1,
#   minimize = TRUE,
#   lower = NULL,
#   upper = NULL,
#   iteration_limit = Inf,
#   seconds_limit = Inf,
#   tolerance_value = 1e-6,
#   tolerance_parameter = 1e-6,
#   tolerance_parameter_norm = function(x, y) sqrt(sum((x - y)^2)),
#   tolerance_history = 1,
#   base_optimizer = Optimizer$new("stats::optim", method = "L-BFGS-B"),
#   verbose = FALSE,
#   hide_warnings = TRUE,
#   add_details = TRUE
# )

## ----himmelblau---------------------------------------------------------------
himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2

## ----himmelblau evaluate------------------------------------------------------
himmelblau(c(3, 2))

## ----visualize_himmelblau, echo = FALSE, message = FALSE, warning = FALSE, fig.align = 'center', out.width = "80%", fig.dim = c(8, 6)----
library("ggplot2")
x <- y <- seq(-5, 5, 0.1)
grid <- expand.grid(x, y)
grid$z <- apply(grid, 1, himmelblau)
ggplot(grid, aes(x = Var1, y = Var2, z = z)) +
  geom_raster(aes(fill = z)) +
  geom_contour(colour = "white", bins = 40) +
  scale_fill_gradient(low = "blue", high = "red") +
  theme_linedraw() +
  labs(
    x = "x",
    y = "y",
    fill = "value",
    title = "Himmelblau function",
    subtitle = "the four local minima are marked in green"
  ) +
  coord_fixed() +
  annotate(
    "Text",
    x = c(3, -2.8, -3.78, 3.58),
    y = c(2, 3.13, -3.28, -1.85),
    label = "X", size = 6, color = "green"
  )

## ----ao himmelblau------------------------------------------------------------
ao(f = himmelblau, initial = c(0, 0))

## ----himmelblau gradient------------------------------------------------------
gradient <- function(x) {
  c(
    4 * x[1] * (x[1]^2 + x[2] - 11) + 2 * (x[1] + x[2]^2 - 7),
    2 * (x[1]^2 + x[2] - 11) + 4 * x[2] * (x[1] + x[2]^2 - 7)
  )
}

## ----ao himmelblau with gradient----------------------------------------------
ao(f = himmelblau, initial = c(0, 0), gradient = gradient, add_details = FALSE)

## ----partition demo-----------------------------------------------------------
process <- ao:::Process$new(
  npar = 10,
  partition = "random",
  new_block_probability = 0.5,
  minimum_block_number = 2
)
process$get_partition()
process$get_partition()

## ----visualize_faithful, echo = FALSE, message = FALSE, warning = FALSE, fig.align = 'center', out.width = "80%", fig.dim = c(8, 6)----
library("ggplot2")
ggplot(datasets::faithful, aes(x = eruptions)) +
  geom_histogram(aes(y = after_stat(density)), bins = 30) +
  xlab("eruption time (min)")

## ----normal mixture-----------------------------------------------------------
normal_mixture_llk <- function(theta, data) {
  mu <- theta[1:2]
  sd <- exp(theta[3:4])
  lambda <- plogis(theta[5])
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}

## ----ao normal mixture--------------------------------------------------------
out <- ao(
  f = normal_mixture_llk,
  initial = runif(5),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE
)
round(out$details, 2)

## ----normal mixture type 2----------------------------------------------------
normal_mixture_llk <- function(data, mu, lsd, llambda) {
  sd <- exp(lsd)
  lambda <- plogis(llambda)
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}

## ----ao normal mixture type 2, results = "hide"-------------------------------
ao(
  f = normal_mixture_llk,
  initial = runif(5),
  target = c("mu", "lsd", "llambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE
)

## ----normal mixture type 3, results = "hide"----------------------------------
normal_mixture_llk <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
ao(
  f = normal_mixture_llk,
  initial = runif(5),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1)
)

## ----normal mixture type 4, warning = FALSE-----------------------------------
normal_mixture_llk <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
out <- ao(
  f = normal_mixture_llk,
  initial = list(runif(5), runif(5)),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = list("random", "random", "random"),
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1)
)
names(out)
out$values

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.align = "center",
  fig.dim = c(8, 6),
  fig.path = "ao-",
  out.width = "80%"
)
library("ao")
set.seed(2)

## ----ao call, eval = FALSE----------------------------------------------------
# ao(
#   f,
#   initial,
#   target = NULL,
#   npar = NULL,
#   gradient = NULL,
#   hessian = NULL,
#   ...,
#   partition = "sequential",
#   new_block_probability = 0.3,
#   minimum_block_number = 1,
#   minimize = TRUE,
#   lower = NULL,
#   upper = NULL,
#   iteration_limit = Inf,
#   seconds_limit = Inf,
#   tolerance_value = 1e-6,
#   tolerance_parameter = 1e-6,
#   tolerance_parameter_norm = function(x, y) sqrt(sum((x - y)^2)),
#   tolerance_history = 1,
#   base_optimizer = optimizeR::Optimizer$new(
#     "stats::optim", method = "L-BFGS-B"
#   ),
#   verbose = FALSE,
#   hide_warnings = TRUE,
#   add_details = TRUE
# )

## ----himmelblau---------------------------------------------------------------
himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2

## ----himmelblau evaluate------------------------------------------------------
himmelblau(c(3, 2))

## ----visualize_himmelblau, echo = FALSE, message = FALSE, warning = FALSE-----
library("ggplot2")
x <- y <- seq(-5, 5, 0.1)
grid <- expand.grid(x, y)
grid$z <- apply(grid, 1, himmelblau)
ggplot(grid, aes(x = Var1, y = Var2, z = z)) +
  geom_raster(aes(fill = z)) +
  geom_contour(colour = "white", bins = 40) +
  scale_fill_gradient(low = "blue", high = "red") +
  theme_linedraw() +
  labs(
    x = "x",
    y = "y",
    fill = "value",
    title = "Himmelblau function",
    subtitle = "the four local minima are marked in green"
  ) +
  coord_fixed() +
  annotate(
    "Text",
    x = c(3, -2.8, -3.78, 3.58),
    y = c(2, 3.13, -3.28, -1.85),
    label = "X", size = 6, color = "green"
  )

## ----ao himmelblau------------------------------------------------------------
ao(f = himmelblau, initial = c(0, 0))

## ----himmelblau gradient------------------------------------------------------
gradient <- function(x) {
  c(
    4 * x[1] * (x[1]^2 + x[2] - 11) + 2 * (x[1] + x[2]^2 - 7),
    2 * (x[1]^2 + x[2] - 11) + 4 * x[2] * (x[1] + x[2]^2 - 7)
  )
}

## ----ao himmelblau with gradient----------------------------------------------
ao(f = himmelblau, initial = c(0, 0), gradient = gradient, add_details = FALSE)

## ----partition demo-----------------------------------------------------------
process <- ao:::Process$new(
  npar = 10,
  partition = "random",
  new_block_probability = 0.5,
  minimum_block_number = 2
)
process$get_partition()
process$get_partition()

## ----visualize_faithful, echo = FALSE, message = FALSE, warning = FALSE-------
library("ggplot2")
ggplot(datasets::faithful, aes(x = eruptions)) +
  geom_histogram(aes(y = after_stat(density)), bins = 30) +
  xlab("eruption time (min)")

## ----normal mixture-----------------------------------------------------------
normal_mixture_llk <- function(theta, data) {
  mu <- theta[1:2]
  sd <- exp(theta[3:4])
  lambda <- plogis(theta[5])
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}

## ----ao normal mixture--------------------------------------------------------
out <- ao(
  f = normal_mixture_llk,
  initial = runif(5),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE
)
round(out$details, 2)

## ----normal mixture type 2----------------------------------------------------
normal_mixture_llk <- function(data, mu, lsd, llambda) {
  sd <- exp(lsd)
  lambda <- plogis(llambda)
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}

## ----ao normal mixture type 2, results = "hide"-------------------------------
ao(
  f = normal_mixture_llk,
  initial = runif(5),
  target = c("mu", "lsd", "llambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE
)

## ----normal mixture type 3, results = "hide"----------------------------------
normal_mixture_llk <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
ao(
  f = normal_mixture_llk,
  initial = runif(5),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1)
)

## ----normal mixture type 4, warning = FALSE-----------------------------------
normal_mixture_llk <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
out <- ao(
  f = normal_mixture_llk,
  initial = list(runif(5), runif(5)),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = list("random", "random", "random"),
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1)
)
names(out)
out$values

#' Split estimate by target
#'
#' @description
#' This helper function splits the solution by target parameters (if provided),
#' which is used for the output.
#'
#' @author Siddhartha Chib
#'
#' @param estimate \[`numeric`\]\cr
#' A parameter vector.
#'
#' @inheritParams ao
#'
#' @return
#' A (named) \code{list}, a partition of `estimate` according to `npar`.

split_by_target <- function(estimate, target = NULL, npar) {
  if (is.null(target)) return(NULL)
  stopifnot(
    is.character(target), is.numeric(estimate),
    length(target) <= length(estimate), is.numeric(npar),
    sum(npar) == length(estimate)
  )
  out <- list()
  idx <- 1L
  for (i in seq_along(target)) {
    k <- as.integer(npar[i])
    out[[target[i]]] <- estimate[idx:(idx + k - 1L)]
    idx <- idx + k
  }
  out
}

#' Generate random partition
#'
#' @description
#' This helper function generates a random parameter partition, which is used
#' for the randomized AO procedure.
#'
#' @author Siddhartha Chib
#'
#' @param x \[`integer`\]\cr
#' The parameter indices.
#'
#' @param p \[`numeric(1)`\]\cr
#' The probability of generating a new block.
#'
#' @param min \[`integer(1)`\]\cr
#' The minimum number of blocks.
#'
#' @return
#' A \code{list}, a random partition of `x`.

generate_random_partition <- function(x, p, min) {
  if (min == length(x)) {
    return(as.list(x))
  }
  x <- sample(x, replace = FALSE)
  n <- length(x)
  y <- sample(0:1, n, replace = TRUE, prob = c(1 - p, p))
  y[1] <- 1
  ind <- which(y == 1)
  if (length(ind) < min) {
    ind <- sort(c(ind, sample(which(y == 0), size = min - length(ind))))
  }
  B <- length(ind)
  blocks <- vector("list", B)
  for (j in seq_len(B)) {
    s <- ind[j]
    e <- if (j < B) ind[j + 1] - 1 else n
    xj <- x[s:e]
    blocks[[j]] <- xj[order(xj)]
  }
  blocks
}

#' Merge optimization results
#'
#' @description
#' This helper function merges the results of multiple AO processes.
#'
#' @param results \[`list`\]\cr
#' A `list` of outputs from \code{\link[ao]{ao}}.
#'
#' @param processes \[`data.frame`\]\cr
#' A `data.frame` describing how the different processes were specified.
#'
#' @inheritParams ao
#'
#' @return
#' A \code{list}, see section "Output value" on the \code{\link[ao]{ao}} page.

merge_results <- function(
    results, minimize = TRUE, add_details = TRUE, processes = data.frame()
  ) {
  values <- vapply(results, `[[`, numeric(1), "value")
  stopping_reasons <- vapply(results, `[[`, character(1), "stopping_reason")
  optimal_process <- ifelse(
    isTRUE(minimize), which.min(values), which.max(values)
  )
  seconds_each <- vapply(results, `[[`, numeric(1), "seconds")
  has_es <- sapply(results, function(x) "estimate_split" %in% names(x)) |> any()
  if (isTRUE(add_details)) {
    details_list <- list()
    for (process in seq_along(results)) {
      details_list[[process]] <- cbind(
        process = process,
        results[[process]][["details"]]
      )
    }
    return(
      c(
        list(
          "estimate" = lapply(results, `[[`, "estimate")[[optimal_process]],
          "estimates" = lapply(results, `[[`, "estimate")
        ),
        if (isTRUE(has_es)) list(
          "estimate_split" = lapply(
            results, `[[`, "estimate_split"
          )[[optimal_process]]
        ),
        list(
          "value" = values[optimal_process],
          "values" = as.list(values),
          "details" = do.call("rbind", details_list),
          "seconds" = sum(seconds_each),
          "seconds_each" = as.list(seconds_each),
          "stopping_reason" = stopping_reasons[optimal_process],
          "stopping_reasons" = as.list(stopping_reasons),
          "processes" = processes
        )
      )
    )
  } else {
    return(
      c(
        list(
          "estimate" = lapply(results, `[[`, "estimate")[[optimal_process]]
        ),
        if (isTRUE(has_es)) list(
          "estimate_split" = lapply(
            results, `[[`, "estimate_split"
          )[[optimal_process]]
        ),
        list(
          "value" = values[optimal_process],
          "seconds" = sum(seconds_each),
          "stopping_reason" = stopping_reasons[optimal_process]
        )
      )
    )
  }
}

test_that("estimate can be split by target", {
  estimate <- 1:5
  target <- c("a", "b", "c")
  npar <- c(2, 2, 1)
  split <- split_by_target(estimate, target, npar)
  checkmate::expect_list(split, types = "integer", len = length(npar))
  expect_equal(split_by_target(estimate, NULL, npar), NULL)
})

test_that("random partition can be generated", {
  min <- 2
  partition <- generate_random_partition(1:5, 0.5, min)
  checkmate::expect_list(partition, types = "integer", min.len = min)
  expect_equal(sort(unlist(partition)), 1:5)
  expect_equal(generate_random_partition(1:5, 0.5, 5), as.list(1:5))
})

test_that("random partition generation is deterministic with a seed", {
  set.seed(123)
  partition_1 <- generate_random_partition(1:10, 0.5, 2)
  set.seed(123)
  partition_2 <- generate_random_partition(1:10, 0.5, 2)
  expect_identical(partition_1, partition_2)
})

#' Process Object
#'
#' @description
#' This object specifies an AO process.
#'
#' @param target \[`character()` | `NULL`\]\cr
#' The name(s) of the argument(s) over which \code{f} is optimized.
#'
#' This can only be \code{numeric} arguments.
#'
#' If `NULL` (default), the first argument of `f` is optimized.
#'
#' @param npar \[`integer()`\]\cr
#' The length(s) of the target argument(s).
#'
#' @param partition \[`character(1)` | `list()`\]\cr
#' Defines the parameter partition. It can be
#'
#' * `"sequential"` for treating each parameter separately,
#' * `"random"` for a random partition in each iteration,
#' * `"none"` for no partition (which is equivalent to joint optimization),
#' * or a `list` of vectors of parameter indices, specifying a custom
#'   partition for the AO process.
#'
#' @param new_block_probability \[`numeric(1)`\]\cr
#' Only relevant if `partition = "random"`.
#'
#' The probability of creating a new parameter block when creating a random
#' partition.
#'
#' Values close to 0 result in larger parameter blocks, values close to 1
#' result in smaller parameter blocks.
#'
#' @param minimum_block_number \[`integer(1)`\]\cr
#' Only relevant if `partition = "random"`.
#'
#' The minimum number of blocks in random partitions.
#'
#' @param verbose \[`logical(1)`\]\cr
#' Print tracing details during the AO process?
#'
#' @param minimize \[`logical(1)`\]\cr
#' Minimize during the AO process?
#'
#' If \code{FALSE}, maximization is performed.
#'
#' @param iteration_limit \[`integer(1)` | `Inf`\]\cr
#' The maximum number of iterations through the parameter partition before
#' the AO process is terminated.
#'
#' Can also be `Inf` for no iteration limit.
#'
#' @param seconds_limit \[`numeric(1)`\]\cr
#' The time limit in seconds before the AO process is terminated.
#'
#' Can also be `Inf` for no time limit.
#'
#' Note that this stopping criterion is only checked *after* a sub-problem is
#' solved and not *within* solving a sub-problem, so the actual process
#' time can exceed this limit.
#'
#' @param tolerance_value \[`numeric(1)`\]\cr
#' A non-negative tolerance value. The AO process terminates
#' if the absolute difference between the current function value and the value
#' from \code{tolerance_history} iterations earlier is smaller than
#' \code{tolerance_value}.
#'
#' Can be `0` for no value threshold.
#'
#' @param tolerance_parameter \[`numeric(1)`\]\cr
#' A non-negative tolerance value. The AO process terminates if
#' the distance between the current estimate and the estimate from
#' \code{tolerance_history} iterations earlier is smaller than
#' \code{tolerance_parameter}.
#'
#' Can be `0` for no parameter threshold.
#'
#' By default, the distance is measured using the Euclidean norm, but another
#' norm can be specified via the \code{tolerance_parameter_norm} field.
#'
#' @param tolerance_parameter_norm \[`function`\]\cr
#' The norm that measures the distance between two estimates. If the distance
#' is smaller than
#' \code{tolerance_parameter}, the AO process is terminated.
#'
#' It must be of the form \code{function(x, y)} for two vector inputs
#' \code{x} and \code{y}, and return a single \code{numeric} value.
#' By default, the Euclidean norm \code{function(x, y) sqrt(sum((x - y)^2))}
#' is used.
#'
#' @param tolerance_history \[`integer(1)`\]\cr
#' The number of iterations to look back to determine whether
#' \code{tolerance_value} or \code{tolerance_parameter} has been reached.
#'
#' @param which_iteration \[`integer()`\]\cr
#' Selects the iteration(s).
#'
#' Can also be \code{NULL} to select all iterations.
#'
#' @param which_block \[`character(1)` | `integer()`\]\cr
#' Selects the parameter block in the partition and can be one of
#'
#' - \code{"first"} for the first parameter block,
#' - \code{"last"} for the last parameter block,
#' - an `integer` vector of parameter indices,
#' - or \code{NULL} for all parameter blocks.
#'
#' @param which_column \[`character()`\]\cr
#' Selects the columns in the `details` part of the output and can be one or
#' more of
#'
#' - `"iteration"`,
#' - `"value"`,
#' - `"parameter"`,
#' - `"block"`,
#' - and `"seconds"`.
#'
#' @param keep_iteration_column \[`logical(1)`\]\cr
#' Keep the column containing the information about the iteration in the output?
#'
#' @param keep_block_columns \[`logical(1)`\]\cr
#' Keep the column containing the information about the active parameter block
#' in the output?
#'
#' @param parameter_type \[`character(1)`\]\cr
#' Selects the parameter type and can be one of
#'
#' - \code{"full"} (default) to get the full parameter vector,
#' - \code{"block"} to get the parameter values for the current block,
#'   i.e., the parameters with the indices `self$block`
#' - \code{"fixed"} to get the parameter values which are currently fixed,
#'   i.e., all except for those with the indices `self$block`
#'
#' @param add_details \[`logical(1)`\]\cr
#' Add details about the AO process to the output?

Process <- R6::R6Class("Process",
   cloneable = FALSE,
   public = list(

   #' @description
   #' Creates a new object of this [R6][R6::R6Class] class.

   initialize = function(
     target = NULL,
     npar = integer(),
     partition = "sequential",
     new_block_probability = 0.3,
     minimum_block_number = 1,
     verbose = FALSE,
     minimize = TRUE,
     iteration_limit = Inf,
     seconds_limit = Inf,
     tolerance_value = 1e-6,
     tolerance_parameter = 1e-6,
     tolerance_parameter_norm = function(x, y) sqrt(sum((x - y)^2)),
     tolerance_history = 1,
     add_details = TRUE
    ) {
     oeli::input_check_response(
       check = checkmate::check_integerish(npar, lower = 0),
       var_name = "npar"
     )
     oeli::input_check_response(
       check = checkmate::check_character(
         target, len = length(npar), unique = TRUE, null.ok = TRUE
        ),
       var_name = "target"
     )
     private$.npar <- npar
     private$.target <- target
     self$partition <- partition
     self$new_block_probability <- new_block_probability
     self$minimum_block_number <- minimum_block_number
     self$verbose <- verbose
     self$minimize <- minimize
     self$iteration_limit <- iteration_limit
     self$seconds_limit <- seconds_limit
     self$tolerance_value <- tolerance_value
     self$tolerance_parameter <- tolerance_parameter
     self$tolerance_parameter_norm <- tolerance_parameter_norm
     self$tolerance_history <- tolerance_history
     self$add_details <- add_details
     invisible(self)
   },

   #' @description
   #' Prints a status message.
   #'
   #' @param message \[`character(1)`\]\cr
   #' A status message.
   #'
   #' @param message_type \[`integer(1)`\]\cr
   #' The message type, one of the following:
   #'
   #' * `1` for `cli::cli_h1()`
   #' * `2` for `cli::cli_h2()`
   #' * `3` for `cli::cli_h3()`
   #' * `4` for `cli::cli_alert_success()`
   #' * `5` for `cli::cli_alert_info()`
   #' * `6` for `cli::cli_alert_warning()`
   #' * `7` for `cli::cli_alert_danger()`
   #' * `8` for `cli::cat_line()`

   print_status = function(message, message_type = 8, verbose = self$verbose) {
     checkmate::assert_string(message, min.chars = 1)
     checkmate::assert_int(message_type, lower = 1, upper = 8)
     checkmate::assert_flag(verbose)
     if (isTRUE(verbose)) {
       switch(message_type,
              cli::cli_h1(message),
              cli::cli_h2(message),
              cli::cli_h3(message),
              cli::cli_alert_success(message),
              cli::cli_alert_info(message),
              cli::cli_alert_warning(message),
              cli::cli_alert_danger(message),
              cli::cat_line(message)
       )
     }
     invisible(self)
   },

   #' @description
   #' Initializes the `details` part of the output.
   #'
   #' @param initial_parameter \[`numeric()`\]\cr
   #' The starting parameter values for the AO process.
   #'
   #' @param initial_value \[`numeric(1)`\]\cr
   #' The function value at the initial parameters.

   initialize_details = function(initial_parameter, initial_value) {
     private$.details <- structure(
       data.frame(
         t(c(0L, initial_value, initial_parameter, rep(0, sum(self$npar)), 0))
       ),
       names = c(
         "iteration", "value", paste0("p", seq_len(sum(self$npar))),
         paste0("b", seq_len(sum(self$npar))), "seconds"
       )
     )
     invisible(self)
   },

   #' @description
   #' Updates the `details` part of the output.
   #'
   #' @param value \[`numeric(1)`\]\cr
   #' The updated function value.
   #'
   #' @param parameter_block \[`numeric()`\]\cr
   #' The updated parameter values for the active parameter block.
   #'
   #' @param seconds \[`numeric(1)`\]\cr
   #' The time in seconds for solving the sub-problem.
   #'
   #' @param error \[`logical(1)`\]\cr
   #' Did solving the sub-problem result in an error?
   #'
   #' @param error_message \[`character(1)`\]\cr
   #' An error message if `error = TRUE`.
   #'
   #' @param block \[`integer()`\]\cr
   #' The currently active parameter block, represented as parameter indices.

   update_details = function(
     value, parameter_block, seconds, error, error_message, block = self$block
   ) {

     ### check inputs
     check_block <- checkmate::check_integerish(
       block,
       unique = TRUE, lower = 1, upper = sum(self$npar)
     )
     check_value <- checkmate::check_number(
       value,
       na.ok = FALSE, null.ok = FALSE, finite = TRUE
     )
     check_parameter_block <- oeli::check_numeric_vector(
       parameter_block,
       any.missing = FALSE, len = length(block)
     )
     check_seconds <- checkmate::check_number(
       seconds,
       na.ok = FALSE, lower = 0, finite = TRUE, null.ok = FALSE
     )
     check_error <- checkmate::check_flag(error)
     check_results <- all(sapply(
       c(
         check_block, check_value, check_parameter_block, check_seconds,
         check_error, isFALSE(error)
       ),
       isTRUE
     ))

     ### error occurred?
     if (isTRUE(error)) {
       private$.error <- TRUE
       private$.error_message <- paste(
         "solving for block [", paste(block, collapse = ","), "] failed:",
         if (checkmate::test_string(error_message, min.chars = 1)) {
           error_message
         } else {
           "unknown reason"
         }
       )
     }

     ### update details
     rows <- nrow(private$.details)
     if (isTRUE(check_results)) {
       parameter <- self$get_parameter_latest("full")
       parameter[block] <- parameter_block
       private$.details[rows + 1, "value"] <- value
       parameter_columns <- which(startsWith(colnames(private$.details), "p"))
       private$.details[rows + 1, parameter_columns] <- parameter
     } else {
       private$.details[rows + 1, ] <- private$.details[rows, ]
     }
     private$.details[rows + 1, "iteration"] <- self$iteration
     private$.details[rows + 1, "seconds"] <- seconds
     block_columns <- which(startsWith(colnames(private$.details), "b"))
     private$.details[rows + 1, block_columns[block]] <- 1
     private$.details[rows + 1, block_columns[-block]] <- 0

     ### return information
     self$print_status("function value:", 8)
     if (self$verbose) cli::cat_print(value)
     self$print_status("parameter:", 8)
     if (self$verbose) cli::cat_print(self$get_parameter_latest("full"))
     invisible(self)
   },

   #' @description
   #' Get a parameter partition.

   get_partition = function() {
     if (checkmate::test_string(self$partition)) {
       switch(
         self$partition,
         "sequential" = as.list(seq_len(sum(self$npar))),
         "random"     = generate_random_partition(
            x = seq_len(sum(self$npar)),
            p = self$new_block_probability,
            min = self$minimum_block_number
          ),
          "none"       = list(seq_len(sum(self$npar)))
       )
     } else {
       self$partition
     }
   },

   #' @description
   #' Get the `details` part of the output.

   get_details = function(
      which_iteration = NULL,
      which_block = NULL,
      which_column = c("iteration", "value", "parameter", "block", "seconds")
    ) {

     ### input checks
     oeli::input_check_response(
       check = checkmate::check_integerish(
         which_iteration,
         lower = 0, null.ok = TRUE, min.len = 1, any.missing = FALSE
       ),
       var_name = "which_iteration"
     )
     oeli::input_check_response(
       check = list(
         checkmate::test_choice(
           which_block, c("first", "last"), null.ok = TRUE
         ),
         checkmate::test_integerish(
           which_block,
           lower = 1, upper = sum(self$npar), unique = TRUE,
           min.len = 1, max.len = sum(self$npar), any.missing = FALSE
         )
       ),
       var_name = "which_block"
     )
     oeli::input_check_response(
       check = checkmate::check_subset(
         which_column, c("iteration", "value", "parameter", "block", "seconds"),
         empty.ok = TRUE
       ),
       var_name = "which_column"
     )

     ### filter details
     details <- private$.details

     # define block columns early (needed for filtering)
     block_columns <- which(startsWith(colnames(details), "b"))

     if (!is.null(which_iteration)) {
       ids <- details$iteration %in% which_iteration
       details <- details[ids, , drop = FALSE]
     }

     if (!is.null(which_block)) {
       if (is.character(which_block)) {
         rows <- oeli::vector_occurrence(details$iteration, which_block)
       } else {
         target <- sort(as.integer(which_block))
         block_details <- details[, block_columns, drop = FALSE]
         rows <- which(apply(block_details, 1, function(x) {
           active <- which(as.numeric(x) == 1)
           identical(active, target)
         }))
       }
     } else {
       rows <- seq_len(nrow(details))
     }

     ### select columns
     iteration_column <- which(colnames(details) == "iteration")
     value_column <- which(colnames(details) == "value")
     parameter_columns <- which(startsWith(colnames(details), "p"))
     block_columns <- which(startsWith(colnames(details), "b"))
     seconds_column <- which(colnames(details) == "seconds")
     columns <- c(
       if ("iteration" %in% which_column) iteration_column,
       if ("value" %in% which_column) value_column,
       if ("parameter" %in% which_column) parameter_columns,
       if ("block" %in% which_column) block_columns,
       if ("seconds" %in% which_column) seconds_column
     )

     ### return details
     details[rows, columns]
   },

   #' @description
   #' Get the function value in different steps of the AO process.

   get_value = function(which_iteration = NULL,
                        which_block = NULL,
                        keep_iteration_column = FALSE,
                        keep_block_columns = FALSE) {
     ### input checks
     oeli::input_check_response(
       check = checkmate::test_flag(keep_iteration_column),
       var_name = "keep_iteration_column"
     )
     oeli::input_check_response(
       check = checkmate::test_flag(keep_block_columns),
       var_name = "keep_block_columns"
     )

     ### return values
     self$get_details(
       which_iteration = which_iteration, which_block = which_block,
       which_column = c(
         "value",
         if (keep_iteration_column) "iteration",
         if (keep_block_columns) "block"
       )
     )
   },

   #' @description
   #' Get the function value in the latest step of the AO process.

   get_value_latest = function() {
     private$.details[nrow(private$.details), "value"]
   },

   #' @description
   #' Get the optimum function value in the AO process.

   get_value_best = function() {
     values <- private$.details[, "value"]
     id_best <- if (self$minimize) {
       which.min(values)
     } else {
       which.max(values)
     }
     private$.details[id_best, "value"]
   },

   #' @description
   #' Get the parameter values in different steps of the AO process.

   get_parameter = function(which_iteration = self$iteration,
                            which_block = NULL,
                            keep_iteration_column = FALSE,
                            keep_block_columns = FALSE) {
     ### input checks
     oeli::input_check_response(
       check = checkmate::test_flag(keep_iteration_column),
       var_name = "keep_iteration_column"
     )
     oeli::input_check_response(
       check = checkmate::test_flag(keep_block_columns),
       var_name = "keep_block_columns"
     )

     ### get parameters
     self$get_details(
       which_iteration = which_iteration, which_block = which_block,
       which_column = c(
         "parameter",
         if (keep_iteration_column) "iteration",
         if (keep_block_columns) "block"
       )
     )
   },

   #' @description
   #' Get the parameter value in the latest step of the AO process.

   get_parameter_latest = function(parameter_type = "full") {

     ### input checks
     oeli::input_check_response(
       check = checkmate::check_choice(
         parameter_type,
         c("full", "block", "fixed")
       ),
       var_name = "parameter_type"
     )

     ### get latest parameter
     details <- private$.details
     parameter_columns <- which(startsWith(colnames(details), "p"))
     latest_parameter <- as.numeric(details[nrow(details), parameter_columns])

     ### filter for 'parameter_type'
     if (parameter_type == "full") {
       return(latest_parameter)
     } else {
       if (parameter_type == "block") {
         return(latest_parameter[self$block])
       } else {
         return(latest_parameter[-self$block])
       }
     }
   },

   #' @description
   #' Get the optimum parameter value in the AO process.

   get_parameter_best = function(parameter_type = "full") {

     ### input checks
     oeli::input_check_response(
       check = checkmate::check_choice(
         parameter_type,
         c("full", "block", "fixed")
       ),
       var_name = "parameter_type"
     )

     ### find best parameter
     details <- private$.details
     id_best <- if (self$minimize) {
       which.min(details[, "value"])
     } else {
       which.max(details[, "value"])
     }
     parameter_columns <- which(startsWith(colnames(details), "p"))
     best_parameter <- as.numeric(details[id_best, parameter_columns])

     ### filter for 'parameter_type'
     if (parameter_type == "full") {
       return(best_parameter)
     } else {
       if (parameter_type == "block") {
         return(best_parameter[self$block])
       } else {
         return(best_parameter[-self$block])
       }
     }
   },

   #' @description
   #' Get the optimization time in seconds in different steps of the AO process.

   get_seconds = function(which_iteration = NULL,
                          which_block = NULL,
                          keep_iteration_column = FALSE,
                          keep_block_columns = FALSE) {

     ### input checks
     oeli::input_check_response(
       check = checkmate::test_flag(keep_iteration_column),
       var_name = "keep_iteration_column"
     )
     oeli::input_check_response(
       check = checkmate::test_flag(keep_block_columns),
       var_name = "keep_block_columns"
     )

     ### get parameters
     self$get_details(
       which_iteration = which_iteration, which_block = which_block,
       which_column = c(
         "seconds",
         if (keep_iteration_column) "iteration",
         if (keep_block_columns) "block"
       )
     )
   },

   #' @description
   #' Get the total optimization time in seconds of the AO process.

   get_seconds_total = function() {
     sum(self$get_seconds(
       which_iteration = NULL, which_block = NULL,
       keep_iteration_column = FALSE, keep_block_columns = FALSE
     ), na.rm = TRUE)
   },

   #' @description
   #' Checks if the AO process can be terminated.

   check_stopping = function() {

     ### check stopping criteria
     stopping <- FALSE
     while (TRUE) {

       ### check error occurred
       if (isTRUE(private$.error)) {
         message <- private$.error_message
         stopping <- TRUE
         break
       }

       ### check iteration limit
       if (self$iteration >= self$iteration_limit) {
         message <- paste("iteration limit of", self$iteration_limit, "reached")
         stopping <- TRUE
         break
       }

       ### check time limit
       if (self$get_seconds_total() >= self$seconds_limit) {
         message <- paste(
           "time limit of", self$seconds_limit, "seconds reached"
         )
         stopping <- TRUE
         break
       }

       ### only check tolerance if at least 'tolerance_history' iterations have
       ### been performed
       if (self$iteration >= self$tolerance_history) {
         ### check value tolerance
         abs_value_change <- abs(
           self$get_value_latest() - self$get_value(
             which_iteration = self$iteration - self$tolerance_history,
             which_block = "first"
           )
         )
         if (abs_value_change < self$tolerance_value) {
           message <- paste(
             "change in function value between", self$tolerance_history,
             ifelse(self$tolerance_history > 1, "iterations", "iteration"),
             "is <", self$tolerance_value
           )
           stopping <- TRUE
           break
         }

         ### check parameter tolerance
         parameter_change <- self$tolerance_parameter_norm(
           self$get_parameter_latest(),
           self$get_parameter(
             which_iteration = self$iteration - self$tolerance_history,
             which_block = "first"
           )
         )
         if (parameter_change < self$tolerance_parameter) {
           message <- paste(
             "change in function parameters between", self$tolerance_history,
             ifelse(self$tolerance_history > 1, "iterations", "iteration"),
             "is <", self$tolerance_parameter
           )
           stopping <- TRUE
           break
         }
       }
       break
     }

     ### decide for stopping
     if (isTRUE(stopping)) {
       private$.stopping_reason <- message
       self$print_status("\n", 8)
       self$print_status(paste("process is terminated:", message), 4)
       self$print_status("\n", 8)
     }
     return(stopping)
   }

 ),

 active = list(

   #' @field npar \[`integer()`\]\cr
   #' The length(s) of the target argument(s).

   npar = function(value) {
     if (missing(value)) {
       private$.npar
     } else {
       cli::cli_abort(
         "Field {.var npar} is read-only",
         call = NULL
       )
     }
   },

   #' @field partition \[`character(1)` | `list()`\]\cr
   #' Defines the parameter partition. It can be
   #'
   #' * `"sequential"` for treating each parameter separately,
   #' * `"random"` for a random partition in each iteration,
   #' * `"none"` for no partition (which is equivalent to joint optimization),
   #' * or a `list` of vectors of parameter indices, specifying a custom
   #'   partition for the AO process.

   partition = function(value) {
     if (missing(value)) {
       private$.partition
     } else {
       if (checkmate::test_string(value)) {
         oeli::input_check_response(
           check = checkmate::check_choice(
             value,
             c("sequential", "random", "none")
           ),
           var_name = "partition"
         )
       } else if (checkmate::test_list(value)) {
         oeli::input_check_response(
           check = checkmate::check_list(value),
           var_name = "partition"
         )

         ### only parameter indices 1,...,'sum(self$npar)' allowed
         oeli::input_check_response(
           check = checkmate::check_integerish(
             unlist(value),
             lower = 1, upper = sum(self$npar), any.missing = FALSE
           ),
           var_name = "partition",
           prefix = "Elements in {.cls list} {.var partition} are bad:"
         )
       } else {
         oeli::input_check_response(
           check = "Bad specification",
           var_name = "partition"
         )
       }
       private$.partition <- value
     }
   },

   #' @field new_block_probability \[`numeric(1)`\]\cr
   #' Only relevant if `partition = "random"`.
   #'
   #' The probability of creating a new parameter block when creating a random
   #' partition.
   #'
   #' Values close to 0 result in larger parameter blocks, values close to 1
   #' result in smaller parameter blocks.

   new_block_probability = function(value) {
     if (missing(value)) {
       private$.new_block_probability
     } else {
       oeli::input_check_response(
         check = checkmate::check_number(value, lower = 0, upper = 1),
         var_name = "new_block_probability"
       )
       private$.new_block_probability <- value
     }
   },

   #' @field minimum_block_number \[`integer(1)`\]\cr
   #' Only relevant if `partition = "random"`.
   #'
   #' The minimum number of blocks in random partitions.

   minimum_block_number = function(value) {
     if (missing(value)) {
       private$.minimum_block_number
     } else {
       oeli::input_check_response(
         check = checkmate::check_int(
           value, lower = 1, upper = sum(self$npar)
         ),
         var_name = "minimum_block_number"
       )
       private$.minimum_block_number <- value
     }
   },

   #' @field verbose \[`logical(1)`\]\cr
   #' Print tracing details during the AO process?

   verbose = function(value) {
     if (missing(value)) {
       private$.verbose
     } else {
       oeli::input_check_response(
         check = checkmate::check_flag(value),
         var_name = "verbose"
       )
       private$.verbose <- value
     }
   },

   #' @field minimize \[`logical(1)`\]\cr
   #' Minimize during the AO process?
   #'
   #' If \code{FALSE}, maximization is performed.

   minimize = function(value) {
     if (missing(value)) {
       private$.minimize
     } else {
       oeli::input_check_response(
         check = checkmate::check_flag(value),
         var_name = "minimize"
       )
       private$.minimize <- value
     }
   },

   #' @field iteration_limit \[`integer(1)` | `Inf`\]\cr
   #' The maximum number of iterations through the parameter partition before
   #' the AO process is terminated.
   #'
   #' Can also be `Inf` for no iteration limit.

   iteration_limit = function(value) {
     if (missing(value)) {
       private$.iteration_limit
     } else {
       oeli::input_check_response(
         check = checkmate::check_number(value, lower = 1, finite = FALSE),
         var_name = "iteration_limit"
       )
       if (is.finite(value)) {
         value <- as.integer(value)
       }
       private$.iteration_limit <- value
     }
   },

   #' @field seconds_limit \[`numeric(1)`\]\cr
   #' The time limit in seconds before the AO process is terminated.
   #'
   #' Can also be `Inf` for no time limit.
   #'
   #' Note that this stopping criterion is only checked *after* a sub-problem is
   #' solved and not *within* solving a sub-problem, so the actual process
   #' time can exceed this limit.

   seconds_limit = function(value) {
     if (missing(value)) {
       private$.seconds_limit
     } else {
       oeli::input_check_response(
         check = checkmate::check_number(
           value,
           lower = 0, finite = FALSE, null.ok = FALSE, na.ok = FALSE
         ),
         var_name = "seconds_limit"
       )
       private$.seconds_limit <- value
     }
   },

   #' @field tolerance_value \[`numeric(1)`\]\cr
   #' A non-negative tolerance value. The AO process terminates
   #' if the absolute difference between the current function value and the
   #' value from \code{tolerance_history} iterations earlier is smaller than
   #' \code{tolerance_value}.
   #'
   #' Can be `0` for no value threshold.

   tolerance_value = function(value) {
     if (missing(value)) {
       private$.tolerance_value
     } else {
       oeli::input_check_response(
         check = checkmate::check_number(value, lower = 0, finite = TRUE),
         var_name = "tolerance_value"
       )
       private$.tolerance_value <- value
     }
   },

   #' @field tolerance_parameter \[`numeric(1)`\]\cr
   #' A non-negative tolerance value. The AO process terminates if
   #' the distance between the current estimate and the estimate from
   #' \code{tolerance_history} iterations earlier is smaller than
   #' \code{tolerance_parameter}.
   #'
   #' Can be `0` for no parameter threshold.
   #'
   #' By default, the distance is measured using the Euclidean norm, but another
   #' norm can be specified via the \code{tolerance_parameter_norm} field.

   tolerance_parameter = function(value) {
     if (missing(value)) {
       private$.tolerance_parameter
     } else {
       oeli::input_check_response(
         check = checkmate::check_number(value, lower = 0, finite = FALSE),
         var_name = "tolerance_parameter"
       )
       private$.tolerance_parameter <- value
     }
   },

   #' @field tolerance_parameter_norm \[`function`\]\cr
   #' The norm that measures the distance between two estimates. If the distance
   #' is smaller than
   #' \code{tolerance_parameter}, the AO process is terminated.
   #'
   #' It must be of the form \code{function(x, y)} for two vector inputs
   #' \code{x} and \code{y}, and return a single \code{numeric} value.
   #' By default, the Euclidean norm \code{function(x, y) sqrt(sum((x - y)^2))}
   #' is used.

   tolerance_parameter_norm = function(value) {
     if (missing(value)) {
       private$.tolerance_parameter_norm[[1]]
     } else {
       oeli::input_check_response(
         check = checkmate::check_function(
           value,
           args = c("x", "y"),
           nargs = 2
         ),
         var_name = "tolerance_parameter_norm"
       )
       private$.tolerance_parameter_norm <- list(value)
     }
   },

   #' @field tolerance_history \[`integer(1)`\]\cr
   #' The number of iterations to look back to determine whether
   #' \code{tolerance_value} or \code{tolerance_parameter} has been reached.

   tolerance_history = function(value) {
     if (missing(value)) {
       private$.tolerance_history
     } else {
       oeli::input_check_response(
         check = checkmate::check_count(value, positive = TRUE),
         var_name = "tolerance_history"
       )
       private$.tolerance_history <- value
     }
   },

   #' @field add_details \[`logical(1)`\]\cr
   #' Add details about the AO process to the output?

   add_details = function(value) {
     if (missing(value)) {
       private$.add_details
     } else {
       oeli::input_check_response(
         check = checkmate::check_flag(value),
         var_name = "add_details"
       )
       private$.add_details <- value
     }
   },

   #' @field iteration \[`integer(1)`\]\cr
   #' The current iteration number.

   iteration = function(value) {
     if (missing(value)) {
       private$.iteration
     } else {
       oeli::input_check_response(
         check = checkmate::check_int(value, lower = 0, na.ok = FALSE),
         var_name = "iteration"
       )
       private$.iteration <- as.integer(value)
       self$print_status(
         paste("iteration", private$.iteration, "of", self$iteration_limit), 2
       )
     }
   },

   #' @field block \[`integer()`\]\cr
   #' The currently active parameter block, represented as parameter indices.

   block = function(value) {
     if (missing(value)) {
       private$.block
     } else {
       oeli::input_check_response(
         check = checkmate::check_integerish(
           value,
           unique = TRUE, lower = 1, upper = sum(self$npar)
         ),
         var_name = "block"
       )
       private$.block <- value
       self$print_status(
         paste("block [", paste(value, collapse = ","), "] : "), 3
       )
     }
   },

   #' @field output \[`list()`, read-only\]\cr
   #' The output of the AO process, which is a \code{list} with the following
   #' elements:
   #'
   #' * \code{estimate} is the parameter vector at termination.
   #' * \code{value} is the function value at termination.
   #' * \code{details} is a `data.frame` with full information about the
   #'   AO process.
   #'   For each iteration (column `iteration`) it contains the function value
   #'   (column `value`), parameter values (columns starting with `p`
   #'   followed by the parameter index), the active parameter block
   #'   (columns starting with `b` followed by the parameter index, where
   #'   `1` stands for a parameter contained in the active parameter block
   #'   and `0` if not), and computation times in seconds (column `seconds`).
   #'   Only available if `add_details = TRUE`.
   #' * \code{seconds} is the overall computation time in seconds.
   #' * \code{stopping_reason} is a message explaining why the AO process
   #'   terminated.

   output = function(value) {
     if (missing(value)) {
       estimate <- self$get_parameter_best("full")
       c(
         list("estimate" = estimate),
         if (!is.null(private$.target)) {
           list("estimate_split" = split_by_target(
             estimate, private$.target, self$npar
           ))
         },
         list("value" = self$get_value_best()),
         if (self$add_details) list("details" = self$get_details()),
         list(
           "seconds" = self$get_seconds_total(),
           "stopping_reason" = private$.stopping_reason
         )
       )
     } else {
       cli::cli_abort(
         "Field {.var output} is read-only",
         call = NULL
       )
     }
   }
 ),

 private = list(

   ### inputs
   .target = character(),
   .npar = integer(),
   .partition = NULL,
   .new_block_probability = numeric(),
   .minimum_block_number = integer(),
   .verbose = logical(),
   .minimize = logical(),
   .iteration_limit = integer(),
   .seconds_limit = numeric(),
   .tolerance_value = numeric(),
   .tolerance_parameter = numeric(),
   .tolerance_history = integer(),

   ### must store 'tolerance_parameter_norm' inside list
   .tolerance_parameter_norm = list(function(x, y) sqrt(sum((x - y)^2))),

   ### results
   .iteration = 0L,
   .block = integer(),
   .details = data.frame(),
   .add_details = TRUE,
   .stopping_reason = "not terminated yet",

   ### error status
   .error = FALSE,
   .error_message = character()
 )
)

#' Alternating Optimization
#'
#' @description
#' Alternating optimization (AO) is an iterative process for optimizing a
#' real-valued function jointly over all its parameters by alternating
#' restricted optimization over parameter partitions.
#'
#' @details
#' ## Multiple processes
#' AO can suffer from local optima. To increase the likelihood of finding a
#' better optimum, you can specify:
#'
#' * multiple starting parameters
#' * multiple parameter partitions
#' * multiple base optimizers
#'
#' Use the `initial`, `partition`, and/or `base_optimizer` arguments to provide
#' a `list` of possible values for each parameter. Each combination of initial
#' values, parameter partitions, and base optimizers will create a separate AO
#' process.
#'
#' ### Output value
#' In the case of multiple processes, the output values refer to the best
#' AO process with respect to function value.
#'
#' If `add_details = TRUE`, the following elements are added:
#'
#' * \code{estimates} is a \code{list} of optimal parameters in each process.
#' * \code{values} is a \code{list} of optimal function values in each process.
#' * \code{details} combines details of the single processes and has an
#'   additional column `process` with an index for the different processes.
#' * \code{seconds_each} gives the computation time in seconds for each process.
#' * \code{stopping_reasons} gives the termination message for each process.
#' * \code{processes} gives details on how the processes were specified.
#'
#' ### Parallel computation
#' By default, processes run sequentially. However, since they are independent,
#' they can be parallelized. To enable parallel computation, use the
#' [`{future}` framework](https://future.futureverse.org/). For example, run the
#' following *before* the `ao()` call:
#' \preformatted{
#' future::plan(future::multisession, workers = 4)
#' }
#'
#' ### Progress updates
#' When using multiple processes, setting `verbose = TRUE` to print tracing
#' details during AO is not supported. However, you can still track the progress
#' using the [`{progressr}` framework](https://progressr.futureverse.org/).
#' For example, run the following *before* the `ao()` call:
#' \preformatted{
#' progressr::handlers(global = TRUE)
#' progressr::handlers(
#'   progressr::handler_progress(":percent :eta :message")
#' )
#' }
#'
#' @param f \[`function`\]\cr
#' A \code{function} to be optimized, returning a single \code{numeric} value.
#'
#' The first argument of \code{f} should be a \code{numeric} of the same length
#' as \code{initial}, optionally followed by any other arguments specified by
#' the \code{...} argument.
#'
#' If \code{f} is optimized over an argument other than the first, or over more
#' than one argument, specify this with the \code{target} argument.
#'
#' @param initial \[`numeric()` | `list()`\]\cr
#' The starting parameter values for the target argument(s).
#'
#' This can also be a `list` of multiple starting parameter values, see details.
#'
#' @param target \[`character()` | `NULL`\]\cr
#' The name(s) of the argument(s) over which \code{f} is optimized.
#'
#' This can only be \code{numeric} arguments.
#'
#' If `NULL` (default), the first argument of `f` is optimized.
#'
#' @param npar \[`integer()`\]\cr
#' The length(s) of the target argument(s).
#'
#' Must be specified if more than one target argument is specified via
#' the `target` argument.
#'
#' Can be `NULL` if there is only one target argument, in which case `npar` is
#' set to `length(initial)`.
#'
#' @param gradient \[`function` | `NULL`\]\cr
#' An optional \code{function} that returns the gradient of \code{f}.
#'
#' The function signature of \code{gradient} must be identical to \code{f}.
#'
#' Ignored if `base_optimizer` does not support custom gradient.
#'
#' @param hessian \[`function` | `NULL`\]\cr
#' An optional \code{function} that returns the Hessian of \code{f}.
#'
#' The function signature of \code{hessian} must be identical to \code{f}.
#'
#' Ignored if `base_optimizer` does not support custom Hessian.
#'
#' @param ...
#' Additional arguments to be passed to \code{f} (and \code{gradient}).
#'
#' @param partition \[`character(1)` | `list()`\]\cr
#' Defines the parameter partition. It can be
#'
#' * `"sequential"` for treating each parameter separately,
#' * `"random"` for a random partition in each iteration,
#' * `"none"` for no partition (which is equivalent to joint optimization),
#' * or a `list` of vectors of parameter indices, specifying a custom
#'   partition for the AO process.
#'
#' This can also be a `list` of multiple partition definitions, see details.
#'
#' @param new_block_probability \[`numeric(1)`\]\cr
#' Only relevant if `partition = "random"`.
#'
#' The probability of creating a new parameter block when creating a random
#' partition.
#'
#' Values close to 0 result in larger parameter blocks, values close to 1
#' result in smaller parameter blocks.
#'
#' @param minimum_block_number \[`integer(1)`\]\cr
#' Only relevant if `partition = "random"`.
#'
#' The minimum number of blocks in random partitions.
#'
#' @param minimize \[`logical(1)`\]\cr
#' Minimize during the AO process?
#'
#' If \code{FALSE}, maximization is performed.
#'
#' @param lower,upper \[`numeric()` | `NULL`\]\cr
#' Optional lower and upper parameter bounds.
#'
#' Ignored if `base_optimizer` does not support parameter bounds.
#'
#' @param iteration_limit \[`integer(1)` | `Inf`\]\cr
#' The maximum number of iterations through the parameter partition before
#' the AO process is terminated.
#'
#' Can also be `Inf` for no iteration limit.
#'
#' @param seconds_limit \[`numeric(1)`\]\cr
#' The time limit in seconds before the AO process is terminated.
#'
#' Can also be `Inf` for no time limit.
#'
#' Note that this stopping criterion is only checked *after* a sub-problem is
#' solved and not *within* solving a sub-problem, so the actual process
#' time can exceed this limit.
#'
#' @param tolerance_value \[`numeric(1)`\]\cr
#' A non-negative tolerance value. The AO process terminates
#' if the absolute difference between the current function value and the value
#' from \code{tolerance_history} iterations earlier is smaller than
#' \code{tolerance_value}.
#'
#' Can be `0` for no value threshold.
#'
#' @param tolerance_parameter \[`numeric(1)`\]\cr
#' A non-negative tolerance value. The AO process terminates if
#' the distance between the current estimate and the estimate from
#' \code{tolerance_history} iterations earlier is smaller than
#' \code{tolerance_parameter}.
#'
#' Can be `0` for no parameter threshold.
#'
#' By default, the distance is measured using the Euclidean norm, but another
#' norm can be specified via the \code{tolerance_parameter_norm} argument.
#'
#' @param tolerance_parameter_norm \[`function`\]\cr
#' The norm that measures the distance between two estimates. If the distance
#' is smaller than
#' \code{tolerance_parameter}, the AO process is terminated.
#'
#' It must be of the form \code{function(x, y)} for two vector inputs
#' \code{x} and \code{y}, and return a single \code{numeric} value.
#' By default, the Euclidean norm \code{function(x, y) sqrt(sum((x - y)^2))}
#' is used.
#'
#' @param tolerance_history \[`integer(1)`\]\cr
#' The number of iterations to look back to determine whether
#' \code{tolerance_value} or \code{tolerance_parameter} has been reached.
#'
#' @param base_optimizer \[`Optimizer` | `list()`\]\cr
#' An \code{Optimizer} object, which can be created via
#' \code{\link[optimizeR]{Optimizer}}. It numerically solves the sub-problems.
#'
#' By default, the \code{\link[stats]{optim}} optimizer with
#' \code{method = "L-BFGS-B"} is used. The default value for the number of
#' iterations within each block is 10.
#'
#' This can also be a `list` of multiple base optimizers, see details.
#'
#' @param verbose \[`logical(1)`\]\cr
#' Print tracing details during the AO process?
#'
#' Not supported when using multiple processes, see details.
#'
#' @param hide_warnings \[`logical(1)`\]\cr
#' Hide warnings during the AO process?
#'
#' @param add_details \[`logical(1)`\]\cr
#' Add details about the AO process to the output?
#'
#' @return
#' A `list` with the following elements:
#'
#' * `estimate` is the parameter vector at termination.
#' * `estimate_split` is `estimate` split by `target` (only when applicable).
#' * `value` is the function value at termination.
#' * `details` is a `data.frame` with information about the AO process:
#'   For each iteration (column `iteration`) it contains the function value
#'   (column `value`), parameter values (columns starting with `p` followed by
#'   the parameter index), the active parameter block (columns starting with `b`
#'   followed by the parameter index, where `1` stands for a parameter contained
#'   in the active parameter block and `0` if not), and computation times in
#'   seconds (column `seconds`). Only available if `add_details = TRUE`.
#' * `seconds` is the overall computation time in seconds.
#' * `stopping_reason` is a message explaining why the AO process terminated.
#'
#' In the case of multiple processes, the output changes slightly, see details.
#'
#' @examples
#' # Example 1: Minimization of Himmelblau's function --------------------------
#'
#' himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
#' ao(f = himmelblau, initial = c(0, 0))
#'
#' # Example 2: Maximization of 2-class Gaussian mixture log-likelihood --------
#'
#' normal_mixture_loglik_uc = function(mu, logsd, eta, data) {
#'   sd <- exp(logsd)
#'   e <- exp(eta[1])
#'   den <- 1 + e
#'   q1 <- e / den
#'   q2 <- 1 / den
#'   l1 <- log(q1) + dnorm(data, mu[1], sd[1], log = TRUE)
#'   l2 <- log(q2) + dnorm(data, mu[2], sd[2], log = TRUE)
#'   m <- pmax(l1, l2)
#'   sum(m + log(exp(l1 - m) + exp(l2 - m)))
#' }
#'
#' set.seed(123)
#'
#' data <- datasets::faithful$eruptions
#'
#' fit <- ao(
#'   f = normal_mixture_loglik_uc,
#'   initial = c(mean(data) + c(-1, 1), rep(log(sd(data)), 2), 0),
#'   target = c("mu", "logsd", "eta"),
#'   npar = c(2, 2, 1),
#'   data = data,
#'   partition = "random",
#'   base_optimizer = optimizeR::Optimizer$new("ucminf::ucminf"),
#'   minimize = FALSE,
#'   add_details = FALSE
#' )
#'
#' (muhat <- fit$estimate_split$mu)
#' (sdhat <- exp(fit$estimate_split$logsd))
#' e <- exp(fit$estimate_split$eta)
#' den <- 1 + e
#' (qhat <- c(e / den, 1 / den))
#'
#' # Example 3: Constrained Optimization in the Setting of Example 2 -----------
#'
#' # target arguments:
#' # - class means mu (2, unrestricted)
#' # - class standard deviations sd (2, must be non-negative)
#' # - class proportion lambda (only 1 for identification, must be in [0, 1])
#'
#' normal_mixture_loglik <- function(mu, sd, lambda, data) {
#'   c1 <- lambda * dnorm(data, mu[1], sd[1])
#'   c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
#'   sum(log(c1 + c2))
#' }
#'
#' set.seed(123)
#'
#' ao(
#'   f = normal_mixture_loglik,
#'   initial = runif(5),
#'   target = c("mu", "sd", "lambda"),
#'   npar = c(2, 2, 1),
#'   data = datasets::faithful$eruptions,
#'   partition = list("sequential", "random", "none"),
#'   minimize = FALSE,
#'   lower = c(-Inf, -Inf, 0, 0, 0),
#'   upper = c(Inf, Inf, Inf, Inf, 1),
#'   add_details = FALSE
#' )
#'
#' @export

ao <- function(
  f,
  initial,
  target = NULL,
  npar = NULL,
  gradient = NULL,
  hessian = NULL,
  ...,
  partition = "sequential",
  new_block_probability = 0.3,
  minimum_block_number = 1,
  minimize = TRUE,
  lower = NULL,
  upper = NULL,
  iteration_limit = Inf,
  seconds_limit = Inf,
  tolerance_value = 1e-6,
  tolerance_parameter = 1e-6,
  tolerance_parameter_norm = function(x, y) sqrt(sum((x - y)^2)),
  tolerance_history = 1,
  base_optimizer = optimizeR::Optimizer$new(
    "stats::optim", method = "L-BFGS-B", control = list("maxit" = 10)
  ),
  verbose = FALSE,
  hide_warnings = TRUE,
  add_details = TRUE
) {

  ### check if required arguments are specified
  oeli::input_check_response(
    check = oeli::check_missing(f),
    var_name = "f"
  )
  oeli::input_check_response(
    check = oeli::check_missing(initial),
    var_name = "initial"
  )

  ### multiple processes? then call ao() for each one
  if (
    is.list(initial) ||
    (is.list(partition) && !checkmate::test_list(partition, "numeric")) ||
    is.list(base_optimizer)
  ) {

    ### build matrix of processes
    if (!is.list(initial)) {
      initial <- list(initial)
    }
    if (!is.list(partition) || checkmate::test_list(partition, "numeric")) {
      partition <- list(partition)
    }
    if (!is.list(base_optimizer)) {
      base_optimizer <- list(base_optimizer)
    }
    processes <- expand.grid(
      initial = initial, partition = partition, base_optimizer = base_optimizer
    )
    nprocesses <- nrow(processes)

    ### run processes (determined by future)
    if (isTRUE(verbose)) {
      verbose <- FALSE
      if (!isTRUE(hide_warnings)) {
        cli::cli_warn("Argument {.var verbose} is set to {.code FALSE}")
      }
    }
    progress_step <- progressr::progressor(steps = nprocesses)
    progress_step(
      paste("running", nprocesses, "AO processes"),
      amount = 0, class = "sticky"
    )
    results <- future.apply::future_lapply(
      seq_len(nprocesses),
      function(process) {
        progress_step(
          paste0("[process ", process, "] started"),
          amount = 0, class = "sticky"
        )
        out <- ao(
          f = f,
          initial = processes[process, "initial"][[1]],
          target = target,
          npar = npar,
          gradient = gradient,
          ...,
          partition = processes[process, "partition"][[1]],
          new_block_probability = new_block_probability,
          minimum_block_number = minimum_block_number,
          minimize = minimize,
          lower = lower,
          upper = upper,
          iteration_limit = iteration_limit,
          seconds_limit = seconds_limit,
          tolerance_value = tolerance_value,
          tolerance_parameter = tolerance_parameter,
          tolerance_parameter_norm = tolerance_parameter_norm,
          tolerance_history = tolerance_history,
          base_optimizer = processes[process, "base_optimizer"][[1]],
          verbose = verbose,
          hide_warnings = hide_warnings,
          add_details = add_details
        )
        progress_step(
          paste0("[process ", process, "] finished"),
          class = "sticky"
        )
        return(out)
      },
      future.seed = TRUE
    )

    ### merge and return results of multiple processes
    return(merge_results(results, minimize, add_details, processes))
  }

  ### input checks and building of objects
  oeli::input_check_response(
    check = oeli::check_numeric_vector(initial, any.missing = FALSE),
    var_name = "initial"
  )
  if (is.null(npar)) {
    ### if npar is not defined, set it to length of initial vector
    npar <- length(initial)
  }
  objective <- Objective$new(f = f, target = target, npar = npar, ...)
  if (!is.null(gradient)) {
    oeli::input_check_response(
      check = checkmate::check_function(gradient),
      var_name = "gradient"
    )
    objective$set_gradient(gradient = gradient, .verbose = isTRUE(verbose))
  }
  if (!is.null(hessian)) {
    oeli::input_check_response(
      check = checkmate::check_function(hessian),
      var_name = "hessian"
    )
    objective$set_hessian(hessian = hessian, .verbose = isTRUE(verbose))
  }
  oeli::input_check_response(
    check = oeli::check_numeric_vector(initial, len = sum(npar)),
    var_name = "initial"
  )
  oeli::input_check_response(
    check = oeli::check_numeric_vector(
      lower, any.missing = FALSE, null.ok = TRUE
    ),
    var_name = "lower"
  )
  if (!is.null(lower)) {
    if (length(lower) == 1) {
      lower <- rep(lower, sum(npar))
    }
    oeli::input_check_response(
      check = oeli::check_numeric_vector(lower, len = sum(npar)),
      var_name = "lower"
    )
    oeli::input_check_response(
      check = if (any(initial < lower)) "Must respect lower limit" else TRUE,
      var_name = "initial"
    )
  }
  oeli::input_check_response(
    check = oeli::check_numeric_vector(
      upper, any.missing = FALSE, null.ok = TRUE
    ),
    var_name = "upper"
  )
  if (!is.null(upper)) {
    if (length(upper) == 1) {
      upper <- rep(upper, sum(npar))
    }
    oeli::input_check_response(
      check = oeli::check_numeric_vector(upper, len = sum(npar)),
      var_name = "upper"
    )
    oeli::input_check_response(
      check = if (any(initial > upper)) "Must respect upper limit" else TRUE,
      var_name = "initial"
    )
  }
  oeli::input_check_response(
    check = checkmate::check_class(base_optimizer, "Optimizer"),
    var_name = "base_optimizer"
  )
  base_optimizer$hide_warnings <- hide_warnings

  ### building AO process
  process <- Process$new(
    target = target,
    npar = npar,
    partition = partition,
    new_block_probability = new_block_probability,
    minimum_block_number = minimum_block_number,
    verbose = verbose,
    minimize = minimize,
    iteration_limit = iteration_limit,
    seconds_limit = seconds_limit,
    tolerance_value = tolerance_value,
    tolerance_parameter = tolerance_parameter,
    tolerance_parameter_norm = tolerance_parameter_norm,
    tolerance_history = tolerance_history,
    add_details = add_details
  )

  ### build sub-problem template
  solve_sub_problem <- function(parameter_block, parameter_fixed, block) {

    block_objective <- Objective$new(
      f = function(p) {
        theta <- numeric(sum(npar))
        theta[block] <- p
        theta[-block] <- parameter_fixed
        objective$evaluate(theta)
      },
      target = "p",
      npar = length(block)
    )

    ### build block gradient and Hessian functions
    if (!is.null(gradient)) {
      block_objective$set_gradient(
        gradient = function(p) {
          theta <- numeric(sum(npar))
          theta[block] <- p
          theta[-block] <- parameter_fixed
          objective$evaluate_gradient(theta)[block]
        },
        .verbose = FALSE
      )
    }
    if (!is.null(hessian)) {
      block_objective$set_hessian(
        hessian = function(p) {
          theta <- numeric(sum(npar))
          theta[block] <- p
          theta[-block] <- parameter_fixed
          objective$evaluate_hessian(theta)[block, block, drop = FALSE]
        },
        .verbose = FALSE
      )
    }

    ### solve sub-problem
    base_optimizer$optimize(
      objective = block_objective,
      initial = parameter_block,
      lower = if (!is.null(lower)) lower[block] else NA,
      upper = if (!is.null(upper)) upper[block] else NA,
      direction = ifelse(process$minimize, "min", "max")
    )
  }

  ### start AO
  process$initialize_details(
    initial_parameter = initial,
    initial_value = objective$evaluate(initial)
  )
  while (TRUE) {
    ### check stopping criteria
    if (process$check_stopping()) {
      break
    } else {
      process$iteration <- process$iteration + 1L
    }

    ### optimize over each parameter block in partition
    for (block in process$get_partition()) {
      process$block <- block

      ### solve sub-problem
      sub_problem_out <- solve_sub_problem(
        parameter_block = process$get_parameter_latest("block"),
        parameter_fixed = process$get_parameter_latest("fixed"),
        block = block
      )

      ### update details
      process$update_details(
        value = sub_problem_out[["value"]],
        parameter_block = sub_problem_out[["parameter"]],
        seconds = sub_problem_out[["seconds"]],
        error = sub_problem_out[["error"]],
        error_message = sub_problem_out[["error_message"]]
      )

    }
  }

  ### return results
  process$output
}

#' @aliases ao-package
#' @inherit ao examples
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom checkmate assert_int
#' @importFrom checkmate assert_number
#' @importFrom checkmate test_class
#' @importFrom checkmate test_flag
#' @importFrom checkmate test_function
#' @importFrom checkmate test_list
#' @importFrom checkmate test_number
#' @importFrom cli cat_line
#' @importFrom cli cli_abort
#' @importFrom cli style_hyperlink
#' @importFrom future.apply future_lapply
#' @importFrom oeli match_arg
#' @importFrom oeli test_numeric_vector
#' @importFrom optimizeR Objective
#' @importFrom progressr progressor
#' @importFrom R6 R6Class
#' @importFrom stats rnorm
#' @importFrom utils packageVersion
## usethis namespace: end
NULL

#' @noRd

.onAttach <- function(lib, pkg) {
  doc_link <- "https://loelschlaeger.de/ao"
  issues_link <- "https://github.com/loelschlaeger/ao/issues"
  msg <- c(
    paste0("This is {ao} ", utils::packageVersion("ao")),
    ", happy alternating optimization!\n",
    "Documentation? ",
    cli::style_hyperlink(doc_link, doc_link), "\n",
    "Any issues? ",
    cli::style_hyperlink(issues_link, issues_link)
  )
  packageStartupMessage(msg)
  invisible()
}

check_ao_output <- function(x, target = NULL, add_details = TRUE) {
  has_target <- !is.null(target)
  names <- c("estimate", "value", "seconds", "stopping_reason")
  if (add_details) {
    names <- oeli::insert_vector_entry(names, "details", 2)
  }
  if (has_target) {
    names <- oeli::insert_vector_entry(names, "estimate_split", 1)
  }
  checkmate::expect_list(x, len = length(names))
  checkmate::expect_names(names(x), identical.to = names)
  if (has_target) {
    checkmate::expect_list(x$estimate_split, len = 3)
    checkmate::expect_names(names(x$estimate_split), identical.to = target)
  }
}

test_that("ao works", {
  himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
  ao_out <- ao(f = himmelblau, initial = c(0, 0))
  check_ao_output(ao_out)
  ao_out <- ao(f = himmelblau, initial = c(0, 0), add_details = FALSE)
  check_ao_output(ao_out, add_details = FALSE)
})

test_that("ao with custom gradient works", {
  himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
  gradient <- function(x) {
    c(
      4 * x[1] * (x[1]^2 + x[2] - 11) + 2 * (x[1] + x[2]^2 - 7),
      2 * (x[1]^2 + x[2] - 11) + 4 * x[2] * (x[1] + x[2]^2 - 7)
    )
  }
  ao_out <- ao(f = himmelblau, initial = c(0, 0), gradient = gradient)
  check_ao_output(ao_out)
  ao_random_out <- ao(
    f = himmelblau, initial = c(0, 0), partition = "random", gradient = gradient
  )
  check_ao_output(ao_random_out)
})

test_that("ao with custom Hessian works", {
  f <- function(x) {
    # print("huhu, my name is f")
    x^2
  }
  g <- function(x) {
    # print("huhu, my name is g")
    2*x
  }
  h <- function(x) {
    # print("huhu, my name is h")
    matrix(2) # important to have matrix(2) instead of 2!
  }
  ao_out <- ao(
    f = f, initial = 10, gradient = g, hessian = h,
    # switch to nlm because stats::optim does not support Hessian
    base_optimizer = optimizeR::Optimizer$new("stats::nlm")
  )
  check_ao_output(ao_out)
})

test_that("ao with additional parameters works", {
  himmelblau <- function(x, a, b, c) {
    (x[1]^a + x[2] - b)^a + (x[1] + x[2]^a - c)^a
  }
  expect_error(
    ao(f = himmelblau, initial = c(0, 0)),
    "Function argument `a` is required but not specified yet."
  )
  ao_out <- ao(f = himmelblau, initial = c(0, 0), a = 2, b = 11, c = 7)
  check_ao_output(ao_out)
})

test_that("ao with targets works", {
  himmelblau <- function(x, a, b, c) {
    (x[1]^a + x[2] - b)^a + (x[1] + x[2]^a - c)^a
  }
  target <- c("x", "b", "c")
  ao_out <- ao(
    f = himmelblau, initial = 1:4, target = target, npar = c(2, 1, 1), a = 1
  )
  check_ao_output(ao_out, target = target)
})

test_that("ao with NULL values for fixed arguments works", {
  f <- function(x, a, b, ind) {
    if (is.null(ind)) {
      (x[1]^2 + x[2] + a)^2 + (x[1] + x[2]^2 + b)^2 + (x[3] - 1)^2
    }
  }
  ao_out <- ao(f = f, initial = c(0, 0, 0), a = -11, b = -7, ind = NULL)
  check_ao_output(ao_out)
})

test_that("ao with a different base optimizer works", {
  himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
  ao_out <- ao(
    f = himmelblau, initial = c(0, 0),
    base_optimizer = optimizeR::Optimizer$new("stats::nlm")
  )
  check_ao_output(ao_out)
})

test_that("ao with custom partition works", {
  f <- function(x) (x[1]^2 + x[2])^2 + (x[1] + x[2]^2)^2 + (x[3] + x[4]^2)^2
  ao_custom <- ao(
    f = f, initial = c(1, 1, 1, 1), partition = list(1, 2, 3:4)
  )
  check_ao_output(ao_custom)
})

test_that("ao with parameter bounds works", {
  rosenbrock <- function(x) (1 - x[1])^2 + (x[2] - x[1]^2)^2
  lower <- 1.5
  upper <- 2.5
  initial <- runif(2, min = lower, max = upper)
  ao_bounds <- ao(
    f = rosenbrock, initial = initial, lower = lower, upper = upper
  )
  check_ao_output(ao_bounds)
  expect_true(all(ao_bounds$estimate <= upper))
  expect_true(all(ao_bounds$estimate >= lower))
})

test_that("ao rejects inconsistent target and npar inputs", {
  f <- function(x, y) sum(x^2) + sum(y^2)
  expect_error(ao(f = f, initial = 1:2, target = c("x", "y")))
})

test_that("ao rejects invalid custom partitions", {
  f <- function(x) sum(x^2)
  expect_error(ao(f = f, initial = 1:2, partition = list(1, 3)))
  expect_error(ao(f = f, initial = 1:2, partition = list(1, NA_integer_)))
})

test_that("ao rejects starting values outside bounds", {
  f <- function(x) sum(x^2)
  expect_error(ao(f = f, initial = 0, lower = 1))
  expect_error(ao(f = f, initial = 2, upper = 1))
})

test_that("multiple ao processes work", {
  himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
  out_multi <- ao(
    f = himmelblau,
    initial = list(c(0, 0), c(1, 1)),
    partition = list("random", list(2, 1), "none", "sequential"),
    base_optimizer = list(
      optimizeR::Optimizer$new("stats::nlm"),
      optimizeR::Optimizer$new("stats::optim")
    ),
    add_details = FALSE
  )
  check_ao_output(out_multi, add_details = FALSE)
})

test_that("multiple ao processes can omit process details", {
  f <- function(x) sum((x - 1)^2)
  out_multi <- ao(
    f = f,
    initial = list(c(0, 0), c(2, 2)),
    partition = "none",
    iteration_limit = 1,
    add_details = FALSE
  )
  check_ao_output(out_multi, add_details = FALSE)
  expect_false(
    any(
      c("details", "estimates", "values", "seconds_each") %in%
        names(out_multi)
    )
  )
})

test_that("ao fails graciously", {
  f <- function(x) x[1]^2 + x[2]^2
  expect_silent(ao(f = f, initial = 1)) # initial is mis-specified
})

test_that("random partition works with ucminf optimizer", {
  skip_if_not_installed("ucminf")
  set.seed(1)

  # simple convex objective in 6D
  f <- function(x) sum((x - 1)^2)
  out <- ao(
    f = f,
    initial = rep(0, 6),
    partition = "random",
    new_block_probability = 0.35,
    minimum_block_number = 2,
    base_optimizer = optimizeR::Optimizer$new("ucminf::ucminf"),
    iteration_limit = 3, # keep it fast for CRAN
    add_details = TRUE
  )
  check_ao_output(out, add_details = TRUE)
  d <- out$details
  bcols <- grep("^b[0-9]+$", names(d), value = TRUE)
  expect_length(bcols, 6)

  # drop init row if present
  d2 <- d[d$iteration > 0, , drop = FALSE]

  # each step should activate at least one parameter and not exceed npar
  active_counts <- rowSums(as.matrix(d2[, bcols, drop = FALSE]) == 1)
  expect_all_true(active_counts >= 1)
  expect_all_true(active_counts <= 6)

  # block indicators must be binary (0/1)
  expect_all_true(unlist(d2[, bcols, drop = FALSE]) %in% c(0, 1))
})

---
title: "Alternating optimization"
description: >
  Learn how to get started with alternating optimization in R.
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Alternating optimization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib
link-citations: true
---

```{r, setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.align = "center",
  fig.dim = c(8, 6),
  fig.path = "ao-",
  out.width = "80%"
)
library("ao")
set.seed(2)
```

The `{ao}` package implements alternating optimization in R. This vignette demonstrates the main workflow and describes the available customization options.

## What is alternating optimization?

Alternating optimization (AO) is an iterative process for optimizing a multivariate function by breaking it into simpler sub-problems. In each step, AO optimizes over one block of parameters while keeping the other blocks fixed, then alternates among the blocks. AO is particularly useful when the sub-problems are easier to solve than the original joint optimization problem, or when the parameters have a natural partition. See @bezdek:2002, @hu:2002, and @bezdek:2003 for more details.

Consider a real-valued **objective** function $f(\mathbf{x}, \mathbf{y})$ where $\mathbf{x}$ and $\mathbf{y}$ are two **blocks** of function parameters, namely a **partition** of the parameters. The AO process can be described as follows:

1. **Initialization**: Start with initial guesses $\mathbf{x}^{(0)}$ and $\mathbf{y}^{(0)}$.

2. **Iterative Steps**: For $k = 0, 1, 2, \dots$
   - **Step 1**: Fix $\mathbf{y} = \mathbf{y}^{(k)}$ and solve the sub-problem $$\mathbf{x}^{(k+1)} = \arg \min_{\mathbf{x}} f(\mathbf{x}, \mathbf{y}^{(k)}).$$
   - **Step 2**: Fix $\mathbf{x} = \mathbf{x}^{(k+1)}$ and solve the sub-problem $$\mathbf{y}^{(k+1)} = \arg \min_{\mathbf{y}} f(\mathbf{x}^{(k+1)}, \mathbf{y}).$$

3. **Convergence**: Repeat the iterative steps until a **convergence criterion** is met, such as when the change in the objective function or the parameters falls below a specified threshold, or when a predefined iteration limit is reached.

The AO process can be

- viewed as a generalization of joint optimization, where the parameter partition is trivial, consisting of the entire parameter vector as a single block,

- also used for maximization problems by simply replacing $\arg \min$ by $\arg \max$ above,

- generalized to more than two parameter blocks, i.e., for $f(\mathbf{x}_1, \mathbf{x}_2, \ldots, \mathbf{x}_n)$, the process involves cycling through each parameter block $\mathbf{x}_1, \mathbf{x}_2, \ldots, \mathbf{x}_n$ and solving the corresponding sub-problems iteratively (the parameter blocks do not necessarily have to be disjoint),

- **randomized** by changing the parameter partition randomly after each iteration, which can further improve the convergence rate and help avoid getting trapped in local optima [@chib:2010],

- run in **multiple processes** for different initial values, parameter partitions, and/or base optimizers.

## How to use the package?

The `{ao}` package provides one user-level function, `ao()`, as the general interface for different AO variants.

### The function call

The `ao()` function call with the default arguments looks as follows:

```{r, ao call, eval = FALSE}
ao(
  f,
  initial,
  target = NULL,
  npar = NULL,
  gradient = NULL,
  hessian = NULL,
  ...,
  partition = "sequential",
  new_block_probability = 0.3,
  minimum_block_number = 1,
  minimize = TRUE,
  lower = NULL,
  upper = NULL,
  iteration_limit = Inf,
  seconds_limit = Inf,
  tolerance_value = 1e-6,
  tolerance_parameter = 1e-6,
  tolerance_parameter_norm = function(x, y) sqrt(sum((x - y)^2)),
  tolerance_history = 1,
  base_optimizer = optimizeR::Optimizer$new(
    "stats::optim", method = "L-BFGS-B"
  ),
  verbose = FALSE,
  hide_warnings = TRUE,
  add_details = TRUE
)
```

The arguments have the following meaning:

- `f`: The objective function to be optimized. By default, `f` is optimized over its first argument. If optimization should target a different argument or multiple arguments, use `npar` and `target`, [see below](#generalized-objective-functions). Additional arguments for `f` can be passed via the `...` argument as usual.

- `initial`: Initial values for the parameters used in the AO process.

- `gradient` and `hessian`: Optional arguments to specify the analytic gradient and/or Hessian of `f`.

- `partition`: Specifies how parameters are partitioned for optimization. Can be one of the following:

    - `"sequential"`: Optimizes each parameter block sequentially. This is similar to [coordinate descent](https://en.wikipedia.org/wiki/Coordinate_descent).

    - `"random"`: Randomly partitions parameters in each iteration.

    - `"none"`: No partitioning; equivalent to joint optimization.

    - Custom partition can be defined using a list of vectors of parameter indices, [see below](#custom-parameter-partition).

- `new_block_probability` and `minimum_block_number` are only relevant if `partition = "random"`. In this case, the former controls the probability for creating a new block when building a random parameter partition, and the latter defines the minimum number of parameter blocks in the partition.

- `minimize`: Set to `TRUE` for minimization (default), or `FALSE` for maximization.

- `lower` and `upper`: Lower and upper limits for constrained optimization.

- `iteration_limit` is the maximum number of AO iterations before termination, while `seconds_limit` is the time limit in seconds. `tolerance_value` and `tolerance_parameter` (in combination with `tolerance_parameter_norm`) specify two other stopping criteria: the current function value or parameter vector must differ from its value `tolerance_history` iterations earlier by less than the corresponding threshold.

- `base_optimizer`: Numerical optimizer used for solving sub-problems, [see below](#base-optimizer).

- Set `verbose` to `TRUE` to print status messages, and `hide_warnings` to `FALSE` to show warnings during the AO process. `add_details = TRUE` adds process details to the output.

### A simple first example

The following is an implementation of the [Himmelblau's function](https://en.wikipedia.org/wiki/Himmelblau%27s_function) $$f(x, y) = (x^2 + y - 11)^2 + (x + y^2 - 7)^2:$$

```{r, himmelblau}
himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
```

This function has four identical local minima, for example in $x = 3$ and $y = 2$:

```{r, himmelblau evaluate}
himmelblau(c(3, 2))
```

```{r, visualize_himmelblau, echo = FALSE, message = FALSE, warning = FALSE}
library("ggplot2")
x <- y <- seq(-5, 5, 0.1)
grid <- expand.grid(x, y)
grid$z <- apply(grid, 1, himmelblau)
ggplot(grid, aes(x = Var1, y = Var2, z = z)) +
  geom_raster(aes(fill = z)) +
  geom_contour(colour = "white", bins = 40) +
  scale_fill_gradient(low = "blue", high = "red") +
  theme_linedraw() +
  labs(
    x = "x",
    y = "y",
    fill = "value",
    title = "Himmelblau function",
    subtitle = "the four local minima are marked in green"
  ) +
  coord_fixed() +
  annotate(
    "Text",
    x = c(3, -2.8, -3.78, 3.58),
    y = c(2, 3.13, -3.28, -1.85),
    label = "X", size = 6, color = "green"
  )
```

Minimizing Himmelblau's function through alternating minimization over $\mathbf{x}$ and $\mathbf{y}$ with initial values $\mathbf{x}^{(0)} = \mathbf{y}^{(0)} = 0$ can be accomplished as follows:

```{r, ao himmelblau}
ao(f = himmelblau, initial = c(0, 0))
```

Here, we see the output of the AO process, which is a `list` that contains the following elements:

- `estimate` is the parameter vector at termination.

- `value` is the function value at termination.

- `details` is a `data.frame` with full information about the process: For each iteration (column `iteration`) it contains the function value (column `value`), parameter values (columns starting with `p` followed by the parameter index), the active parameter block (columns starting with `b` followed by the parameter index, where `1` stands for a parameter contained in the active parameter block and `0` if not), and computation times in seconds (column `seconds`).

- `seconds` is the overall computation time in seconds.

- `stopping_reason` is a message explaining why the process terminated.

### Using the analytic gradient

For Himmelblau's function, the analytic gradient is:

```{r, himmelblau gradient}
gradient <- function(x) {
  c(
    4 * x[1] * (x[1]^2 + x[2] - 11) + 2 * (x[1] + x[2]^2 - 7),
    2 * (x[1]^2 + x[2] - 11) + 4 * x[2] * (x[1] + x[2]^2 - 7)
  )
}
```

The gradient function is used by `ao()` if provided via the `gradient` argument as follows:

```{r, ao himmelblau with gradient}
ao(f = himmelblau, initial = c(0, 0), gradient = gradient, add_details = FALSE)
```

In higher-dimensional problems, using the analytic gradient can improve both the speed and stability of the process. An analytic Hessian can be supplied in the same way.

### Random parameter partitions

Another AO variant uses a new random partition of the parameters in every iteration. This can improve convergence and help prevent the process from getting stuck in local optima. It is activated by setting `partition = "random"`. The randomness can be adjusted using two parameters:

- `new_block_probability` determines the probability of creating a new block when building a new partition. Its value ranges from `0` (no blocks are created) to `1` (each parameter is a single block).

- `minimum_block_number` sets the minimum number of parameter blocks for random partitions. Here, it is configured as `2` to avoid generating trivial partitions.

Random partitions are generated as follows:^[`Process` is an internal R6 object [@chang:2022], which users typically do not need to interact with.]

```{r, partition demo}
process <- ao:::Process$new(
  npar = 10,
  partition = "random",
  new_block_probability = 0.5,
  minimum_block_number = 2
)
process$get_partition()
process$get_partition()
```

As an example of AO with random partitions, consider fitting a two-class Gaussian mixture model via maximizing the model's log-likelihood function

$$\ell(\boldsymbol{\theta}) = \sum_{i=1}^n \log\Big( \lambda \phi_{\mu_1, \sigma_1^2}(x_i) + (1-\lambda)\phi_{\mu_2,\sigma_2^2} (x_i) \Big),$$

where the sum is over all observations $x_1, \dots, x_n$, $\phi_{\mu_1, \sigma_1^2}$ and $\phi_{\mu_2, \sigma_2^2}$ denote the normal density for the first and second cluster, respectively, and $\lambda$ is the mixing proportion. The parameter vector to be estimated is $\boldsymbol{\theta} = (\mu_1, \mu_2, \sigma_1, \sigma_2, \lambda)$. Because there is no closed-form solution for the maximum likelihood estimator $\boldsymbol{\theta}^* = \arg\max_{\boldsymbol{\theta}} \ell(\boldsymbol{\theta})$, we use numerical optimization to find the optimum. The model is fitted to the following data:^[The `faithful` data set contains information about eruption times (`eruptions`) of the Old Faithful geyser in Yellowstone National Park, Wyoming, USA. The data histogram suggests two clusters with short and long eruption times, respectively. For both clusters, we assume a normal distribution, such that we model the overall eruption times with a mixture of two Gaussian densities.]

```{r, visualize_faithful, echo = FALSE, message = FALSE, warning = FALSE}
library("ggplot2")
ggplot(datasets::faithful, aes(x = eruptions)) +
  geom_histogram(aes(y = after_stat(density)), bins = 30) +
  xlab("eruption time (min)")
```

The following function calculates the log-likelihood value given the parameter vector `theta` and the observation vector `data`:^[We restrict the standard deviations `sd` to be positive (via the exponential transformation) and `lambda` to be between 0 and 1 (via the logit transformation).]

```{r, normal mixture}
normal_mixture_llk <- function(theta, data) {
  mu <- theta[1:2]
  sd <- exp(theta[3:4])
  lambda <- plogis(theta[5])
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
```

The `ao()` call for performing alternating *maximization* with random partitions looks as follows, where we simplified the output for brevity:

```{r, ao normal mixture}
out <- ao(
  f = normal_mixture_llk,
  initial = runif(5),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE
)
round(out$details, 2)
```

### More flexibility

The `{ao}` package offers additional flexibility for performing AO.^[Missing functionality? Please let us know via an [issue on GitHub](https://github.com/loelschlaeger/ao/issues/new?assignees=&labels=future&projects=&template=suggestion.md).]

#### Generalized objective functions

Optimizers in R generally require that the objective function has a single target argument in the first position, but `{ao}` can optimize over another argument or over multiple arguments. For example, suppose the `normal_mixture_llk()` function above has the following form and should be optimized over the parameters `mu`, `lsd`, and `llambda`:

```{r, normal mixture type 2}
normal_mixture_llk <- function(data, mu, lsd, llambda) {
  sd <- exp(lsd)
  lambda <- plogis(llambda)
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
```

In `ao()`, this scenario can be specified by setting

- `target = c("mu", "lsd", "llambda")` (the names of the target arguments)

- and `npar = c(2, 2, 1)` (the lengths of the target arguments):

```{r, ao normal mixture type 2, results = "hide"}
ao(
  f = normal_mixture_llk,
  initial = runif(5),
  target = c("mu", "lsd", "llambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE
)
```

#### Parameter bounds

Instead of using parameter transformations in the `normal_mixture_llk()` function above, parameter bounds can be specified via the arguments `lower` and `upper`, each of which can be a single number (a common bound for all parameters) or a vector of parameter-specific bounds. Therefore, a more straightforward implementation of the mixture example would be:

```{r, normal mixture type 3, results = "hide"}
normal_mixture_llk <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
ao(
  f = normal_mixture_llk,
  initial = runif(5),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1)
)
```

#### Custom parameter partition

Say the parameters of the Gaussian mixture model are supposed to be grouped by type:

$$\mathbf{x}_1 = (\mu_1, \mu_2),\ \mathbf{x}_2 = (\sigma_1, \sigma_2),\ \mathbf{x}_3 = (\lambda).$$

In `ao()`, custom parameter partitions can be specified by setting `partition = list(1:2, 3:4, 5)`, i.e. by defining a `list` where each element corresponds to a parameter block, containing a vector of parameter indices. Parameter indices can be members of any number of blocks.

#### Stopping criteria

Currently, four different stopping criteria for the AO process are implemented:

1. a predefined iteration limit is exceeded (via the `iteration_limit` argument)

2. a predefined time limit is exceeded (via the `seconds_limit` argument)

3. the absolute change in the function value in comparison to the last iteration falls below a predefined threshold (via the `tolerance_value` argument)

4. the change in parameters compared with the last iteration falls below a predefined threshold (via the `tolerance_parameter` argument, where the parameter distance is computed via the norm specified as `tolerance_parameter_norm`)

Any number of stopping criteria can be activated or deactivated^[Stopping criteria of the AO process can be deactivated by setting `iteration_limit = Inf`, `seconds_limit = Inf`, `tolerance_value = 0`, or `tolerance_parameter = 0`.], and the final output identifies the criterion that caused termination.

#### Base optimizer

By default, the L-BFGS-B algorithm [@byrd:1995] implemented in `stats::optim` is used to solve the sub-problems numerically. Any other optimizer can be selected with the `base_optimizer` argument. Such an optimizer must be defined through the framework provided by the `{optimizeR}` package; see [its documentation](https://loelschlaeger.de/optimizeR/) for details. For example, the `stats::nlm` optimizer can be selected by setting `base_optimizer = optimizeR::Optimizer$new("stats::nlm")`.

#### Multiple processes
AO can suffer from local optima. To increase the likelihood of finding a better optimum, users can specify

- multiple starting parameters,

- multiple parameter partitions,

- multiple base optimizers.

Use the `initial`, `partition`, and/or `base_optimizer` arguments to provide a `list` of possible values for each parameter. Each combination of initial values, parameter partitions, and base optimizers will create a separate AO process:

```{r, normal mixture type 4, warning = FALSE}
normal_mixture_llk <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
out <- ao(
  f = normal_mixture_llk,
  initial = list(runif(5), runif(5)),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = list("random", "random", "random"),
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1)
)
names(out)
out$values
```

In the case of multiple processes, the output provides information for the best process (with respect to the function value) as well as information for each process.

By default, processes run sequentially. However, since they are independent of each other, they can be parallelized. For parallel computation, `{ao}` supports the [`{future}` framework](https://future.futureverse.org/). For example, run the following *before* the `ao()` call:

```r
future::plan(future::multisession, workers = 4)
```

When using multiple processes, setting `verbose = TRUE` to print tracing details during AO is not supported. However, process progress can still be tracked using the [`{progressr}` framework](https://progressr.futureverse.org/). For example, run the following *before* the `ao()` call:

```r
progressr::handlers(global = TRUE)
progressr::handlers(
  progressr::handler_progress(":percent :eta :message")
)
```

## References

---
title: "Alternating optimization"
description: >
  Learn how to get started with alternating optimization in R.
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Alternating optimization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib
link-citations: true
---

```{r, setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.path = "ao-"
)
library("ao")
set.seed(2)
```

The `{ao}` package implements alternating optimization in R. This vignette demonstrates how to use the package and describes the available customization options.

## What is alternating optimization?

Alternating optimization (AO) is an iterative process for optimizing a multivariate function by breaking it down into simpler sub-problems. It involves optimizing over one block of function parameters while keeping the others fixed, and then alternating this process among the parameter blocks. AO is particularly useful when the sub-problems are easier to solve than the original joint optimization problem, or when there is a natural partitioning of the parameters. See @bezdek:2002, @hu:2002, and @bezdek:2003 for more details.

Consider a real-valued **objective** function $f(\mathbf{x}, \mathbf{y})$ where $\mathbf{x}$ and $\mathbf{y}$ are two **blocks** of function parameters, namely a **partition** of the parameters. The AO process can be described as follows:

1. **Initialization**: Start with initial guesses $\mathbf{x}^{(0)}$ and $\mathbf{y}^{(0)}$.

2. **Iterative Steps**: For $k = 0, 1, 2, \dots$
   - **Step 1**: Fix $\mathbf{y} = \mathbf{y}^{(k)}$ and solve the sub-problem $$\mathbf{x}^{(k+1)} = \arg \min_{\mathbf{x}} f(\mathbf{x}, \mathbf{y}^{(k)}).$$
   - **Step 2**: Fix $\mathbf{x} = \mathbf{x}^{(k+1)}$ and solve the sub-problem $$\mathbf{y}^{(k+1)} = \arg \min_{\mathbf{y}} f(\mathbf{x}^{(k+1)}, \mathbf{y}).$$

3. **Convergence**: Repeat the iterative steps until a **convergence criterion** is met, such as when the change in the objective function or the parameters falls below a specified threshold, or when a pre-defined iteration limit is reached.

The AO process can be

- viewed as a generalization of joint optimization, where the parameter partition is trivial, consisting of the entire parameter vector as a single block,

- also used for maximization problems by simply replacing $\arg \min$ by $\arg \max$ above,

- generalized to more than two parameter blocks, i.e., for $f(\mathbf{x}_1, \mathbf{x}_2, \ldots, \mathbf{x}_n)$, the process involves cycling through each parameter block $\mathbf{x}_1, \mathbf{x}_2, \ldots, \mathbf{x}_n$ and solving the corresponding sub-problems iteratively (the parameter blocks do not necessarily have to be disjoint),

- **randomized** by changing the parameter partition randomly after each iteration, which can further improve the convergence rate and help avoid getting trapped in local optima [@chib:2010],

- run in **multiple processes** for different initial values, parameter partitions, and/or base optimizers.

## How to use the package?

The `{ao}` package provides a single user-level function, `ao()`, which serves as a general interface for performing various variants of AO.

### The function call

The `ao()` function call with the default arguments looks as follows:

```{r, ao call, eval = FALSE}
ao(
  f,
  initial,
  target = NULL,
  npar = NULL,
  gradient = NULL,
  hessian = NULL,
  ...,
  partition = "sequential",
  new_block_probability = 0.3,
  minimum_block_number = 1,
  minimize = TRUE,
  lower = NULL,
  upper = NULL,
  iteration_limit = Inf,
  seconds_limit = Inf,
  tolerance_value = 1e-6,
  tolerance_parameter = 1e-6,
  tolerance_parameter_norm = function(x, y) sqrt(sum((x - y)^2)),
  tolerance_history = 1,
  base_optimizer = Optimizer$new("stats::optim", method = "L-BFGS-B"),
  verbose = FALSE,
  hide_warnings = TRUE,
  add_details = TRUE
)
```

The arguments have the following meaning:

- `f`: The objective function to be optimized. By default, `f` is optimized over its first argument. If optimization should target a different argument or multiple arguments, use `npar` and `target`, [see below](#generalized-objective-functions). Additional arguments for `f` can be passed via the `...` argument as usual.

- `initial`: Initial values for the parameters used in the AO process.

- `gradient` and `hessian`: Optional arguments to specify the analytical gradient and/or Hessian of `f`.

- `partition`: Specifies how parameters are partitioned for optimization. Can be one of the following:

    - `"sequential"`: Optimizes each parameter block sequentially. This is similar to [coordinate descent](https://en.wikipedia.org/wiki/Coordinate_descent).

    - `"random"`: Randomly partitions parameters in each iteration.

    - `"none"`: No partitioning; equivalent to joint optimization.

    - Custom partition can be defined using a list of vectors of parameter indices, [see below](#custom-parameter-partition).

- `new_block_probability` and `minimum_block_number` are only relevant if `partition = "random"`. In this case, the former controls the probability for creating a new block when building a random parameter partition, and the latter defines the minimum number of parameter blocks in the partition.

- `minimize`: Set to `TRUE` for minimization (default), or `FALSE` for maximization.

- `lower` and `upper`: Lower and upper limits for constrained optimization.

- `iteration_limit` is the maximum number of AO iterations before termination, while `seconds_limit` is the time limit in seconds. `tolerance_value` and `tolerance_parameter` (in combination with `tolerance_parameter_norm`) specify two other stopping criteria, namely when the difference between the current function value or the current parameter vector and the one before `tolerance_history` iterations, respectively, becomes smaller than these thresholds.

- `base_optimizer`: Numerical optimizer used for solving sub-problems, [see below](#base-optimizer).

- Set `verbose` to `TRUE` to print status messages, and `hide_warnings` to `FALSE` to show warning messages during the AO process. `add_details = TRUE` adds additional details about the AO process to the output.

### A simple first example

The following is an implementation of the [Himmelblau's function](https://en.wikipedia.org/wiki/Himmelblau%27s_function) $$f(x, y) = (x^2 + y - 11)^2 + (x + y^2 - 7)^2:$$

```{r, himmelblau}
himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
```

This function has four identical local minima, for example in $x = 3$ and $y = 2$:

```{r, himmelblau evaluate}
himmelblau(c(3, 2))
```

```{r, visualize_himmelblau, echo = FALSE, message = FALSE, warning = FALSE, fig.align = 'center', out.width = "80%", fig.dim = c(8, 6)}
library("ggplot2")
x <- y <- seq(-5, 5, 0.1)
grid <- expand.grid(x, y)
grid$z <- apply(grid, 1, himmelblau)
ggplot(grid, aes(x = Var1, y = Var2, z = z)) +
  geom_raster(aes(fill = z)) +
  geom_contour(colour = "white", bins = 40) +
  scale_fill_gradient(low = "blue", high = "red") +
  theme_linedraw() +
  labs(
    x = "x",
    y = "y",
    fill = "value",
    title = "Himmelblau function",
    subtitle = "the four local minima are marked in green"
  ) +
  coord_fixed() +
  annotate(
    "Text",
    x = c(3, -2.8, -3.78, 3.58),
    y = c(2, 3.13, -3.28, -1.85),
    label = "X", size = 6, color = "green"
  )
```

Minimizing Himmelblau's function through alternating minimization over $\mathbf{x}$ and $\mathbf{y}$ with initial values $\mathbf{x}^{(0)} = \mathbf{y}^{(0)} = 0$ can be accomplished as follows:

```{r, ao himmelblau}
ao(f = himmelblau, initial = c(0, 0))
```

Here, we see the output of the AO process, which is a `list` that contains the following elements:

- `estimate` is the parameter vector at termination.

- `value` is the function value at termination.

- `details` is a `data.frame` with full information about the process: For each iteration (column `iteration`) it contains the function value (column `value`), parameter values (columns starting with `p` followed by the parameter index), the active parameter block (columns starting with `b` followed by the parameter index, where `1` stands for a parameter contained in the active parameter block and `0` if not), and computation times in seconds (column `seconds`).

- `seconds` is the overall computation time in seconds.

- `stopping_reason` is a message why the process has terminated.

### Using the analytical gradient

For the Himmelblau's function, it is straightforward to define the analytical gradient as follows:

```{r, himmelblau gradient}
gradient <- function(x) {
  c(
    4 * x[1] * (x[1]^2 + x[2] - 11) + 2 * (x[1] + x[2]^2 - 7),
    2 * (x[1]^2 + x[2] - 11) + 4 * x[2] * (x[1] + x[2]^2 - 7)
  )
}
```

The gradient function is used by `ao()` if provided via the `gradient` argument as follows:

```{r, ao himmelblau with gradient}
ao(f = himmelblau, initial = c(0, 0), gradient = gradient, add_details = FALSE)
```

In scenarios involving higher dimensions, utilizing the analytical gradient can notably improve both the speed and stability of the process. The analytical Hessian can be utilized analogously. 

### Random parameter partitions

Another version of the AO process involves using a new, random partition of the parameters in every iteration. This approach can enhance the convergence rate and prevent being stuck in local optima. It is activated by setting `partition = "random"`. The randomness can be adjusted using two parameters:

- `new_block_probability` determines the probability for creating a new block when building a new partition. Its value ranges from `0` (no blocks are created) to `1` (each parameter is a single block). 

- `minimum_block_number` sets the minimum number of parameter blocks for random partitions. Here, it is configured as `2` to avoid generating trivial partitions.

The random partitions are build as follows:^[`Process` is an internal R6 object [@chang:2022], which users typically do not need to interact with.]

```{r, partition demo}
process <- ao:::Process$new(
  npar = 10,
  partition = "random",
  new_block_probability = 0.5,
  minimum_block_number = 2
)
process$get_partition()
process$get_partition()
```

As an example of AO with random partitions, consider fitting a two-class Gaussian mixture model via maximizing the model's log-likelihood function

$$\ell(\boldsymbol{\theta}) = \sum_{i=1}^n \log\Big( \lambda \phi_{\mu_1, \sigma_1^2}(x_i) + (1-\lambda)\phi_{\mu_2,\sigma_2^2} (x_i) \Big),$$

where the sum goes over all observations $x_1, \dots, x_n$, $\phi_{\mu_1, \sigma_1^2}$ and $\phi_{\mu_2, \sigma_2^2}$ denote the normal density for the first and second cluster, respectively, and $\lambda$ is the mixing proportion. The parameter vector to be estimated is $\boldsymbol{\theta} = (\mu_1, \mu_2, \sigma_1, \sigma_2, \lambda)$. As there exists no closed-form solution for the maximum likelihood estimator $\boldsymbol{\theta}^* = \arg\max_{\boldsymbol{\theta}} \ell(\boldsymbol{\theta})$, we apply numerical optimization to find the function optimum. The model is fitted to the following data:^[The `faithful` data set contains information about eruption times (`eruptions`) of the Old Faithful geyser in Yellowstone National Park, Wyoming, USA. The data histogram hints at two clusters with short and long eruption times, respectively. For both clusters, we assume a normal distribution, such that we consider a mixture of two Gaussian densities for modeling the overall eruption times.]

```{r, visualize_faithful, echo = FALSE, message = FALSE, warning = FALSE, fig.align = 'center', out.width = "80%", fig.dim = c(8, 6)}
library("ggplot2")
ggplot(datasets::faithful, aes(x = eruptions)) +
  geom_histogram(aes(y = after_stat(density)), bins = 30) +
  xlab("eruption time (min)")
```

The following function calculates the log-likelihood value given the parameter vector `theta` and the observation vector `data`:^[We restrict the standard deviations `sd` to be positive (via the exponential transformation) and `lambda` to be between 0 and 1 (via the logit transformation).]

```{r, normal mixture}
normal_mixture_llk <- function(theta, data) {
  mu <- theta[1:2]
  sd <- exp(theta[3:4])
  lambda <- plogis(theta[5])
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
```

The `ao()` call for performing alternating *maximization* with random partitions looks as follows, where we simplified the output for brevity:

```{r, ao normal mixture}
out <- ao(
  f = normal_mixture_llk,
  initial = runif(5),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE
)
round(out$details, 2)
```

### More flexibility

The `{ao}` package offers some flexibility for performing AO.^[Do you miss a functionality? Please let us know via an [issue on GitHub](https://github.com/loelschlaeger/ao/issues/new?assignees=&labels=future&projects=&template=suggestion.md).]

#### Generalized objective functions

Optimizers in R generally require that the objective function has a single target argument which must be in the first position, but `{ao}` allows for optimization over an argument other than the first, or more than one argument. For example, say, the `normal_mixture_llk` function above has the following form and is supposed to be optimized over the parameters `mu`, `lsd`, and `llambda`:

```{r, normal mixture type 2}
normal_mixture_llk <- function(data, mu, lsd, llambda) {
  sd <- exp(lsd)
  lambda <- plogis(llambda)
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
```

In `ao()`, this scenario can be specified by setting

- `target = c("mu", "lsd", "llambda")` (the names of the target arguments)

- and `npar = c(2, 2, 1)` (the lengths of the target arguments):

```{r, ao normal mixture type 2, results = "hide"}
ao(
  f = normal_mixture_llk,
  initial = runif(5),
  target = c("mu", "lsd", "llambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE
)
```

#### Parameter bounds

Instead of using parameter transformations in the `normal_mixture_llk()` function above, parameter bounds can be specified via the arguments `lower` and `upper`, where both can either be a single number (a common bound for all parameters) or a vector of specific bounds per parameter. Therefore, an more straightforward implementation of the mixture example would be:

```{r, normal mixture type 3, results = "hide"}
normal_mixture_llk <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
ao(
  f = normal_mixture_llk,
  initial = runif(5),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = "random",
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1)
)
```

#### Custom parameter partition

Say the parameters of the Gaussian mixture model are supposed to be grouped by type:

$$\mathbf{x}_1 = (\mu_1, \mu_2),\ \mathbf{x}_2 = (\sigma_1, \sigma_2),\ \mathbf{x}_3 = (\lambda).$$

In `ao()`, custom parameter partitions can be specified by setting `partition = list(1:2, 3:4, 5)`, i.e. by defining a `list` where each element corresponds to a parameter block, containing a vector of parameter indices. Parameter indices can be members of any number of blocks.

#### Stopping criteria

Currently, four different stopping criteria for the AO process are implemented:

1. a predefined iteration limit is exceeded (via the `iteration_limit` argument)

2. a predefined time limit is exceeded (via the `seconds_limit` argument)

3. the absolute change in the function value in comparison to the last iteration falls below a predefined threshold (via the `tolerance_value` argument)

4. the change in parameters in comparison to the last iteration falls below a predefined threshold (via the `tolerance_parameter` argument, where the parameter distance is computed via the norm specified as `tolerance_parameter_norm`)

Any number of stopping criteria can be activated or deactivated^[Stopping criteria of the AO process can be deactivated by setting `iteration_limit = Inf`, `seconds_limit = Inf`, `tolerance_value = 0`, or `tolerance_parameter = 0`.], and the final output contains information about the criterium that caused termination.

#### Base optimizer

By default, the L-BFGS-B algorithm [@byrd:1995] implemented in `stats::optim` is used for solving the sub-problems numerically. However, any other optimizer can be selected by specifying the `base_optimizer` argument. Such an optimizer must be defined through the framework provided by the `{optimizeR}` package, see [its documentation](https://loelschlaeger.de/optimizeR/) for details. For example, the `stats::nlm` optimizer can be selected by setting `base_optimizer = Optimizer$new("stats::nlm")`.

#### Multiple processes
AO can suffer from local optima. To increase the likelihood of reaching the global optimum, users can specify

- multiple starting parameters,

- multiple parameter partitions,

- multiple base optimizers.

Use the `initial`, `partition`, and/or `base_optimizer` arguments to provide a `list` of possible values for each parameter. Each combination of initial values, parameter partitions, and base optimizers will create a separate AO process:

```{r, normal mixture type 4, warning = FALSE}
normal_mixture_llk <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}
out <- ao(
  f = normal_mixture_llk,
  initial = list(runif(5), runif(5)),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = list("random", "random", "random"),
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1)
)
names(out)
out$values
```

In the case of multiple processes, the output provides information for the best process (with respect to the function value) as well as information on every single process.

By default, processes run sequentially. However, since they are independent of each other, they can be parallelized. For parallel computation, `{ao}` supports the [`{future}` framework](https://future.futureverse.org/). For example, run the following *before* the `ao()` call:

```r
future::plan(future::multisession, workers = 4)
```

When using multiple processes, setting `verbose = TRUE` to print tracing details during AO is not supported. However, progress of processes can still be tracked using the [`{progressr}` framework](https://progressr.futureverse.org/). For example, run the following *before* the `ao()` call:

```r
progressr::handlers(global = TRUE)
progressr::handlers(
  progressr::handler_progress(":percent :eta :message")
)
```

## References

---
title: "Alternating optimization"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Alternating optimization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib
link-citations: true
---

```{r, setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.path = "ao-"
)
```

## Background

### What is alternating optimization?

Alternating optimization is an iterative procedure for optimizing some function jointly over all parameters by alternating restricted optimization over individual parameter subsets.

More precisely, consider minimizing or maximizing $f:\mathbb{R}^n \to \mathbb{R}$ over the set of feasible points $x \in X \subseteq \mathbb{R}^n$. The underlying algorithm of alternating optimization is as follows:

1. Assign an initial value for $x$.

2. Optimize $f$ with respect to a subset of parameters $\tilde{x}$ while holding the other parameters constant. Note that alternating optimization is a generalization of joint optimization, where the only parameter subset would be the whole set of parameters.

3. Replace the values in $x$ by the optimal values for $\tilde{x}$ found in step 2.

4. Repeat from step 2 with another parameter subset.

5. Stop when the process has converged or reached the iteration limit.

### When is alternating optimization a good idea?

- When the joint optimization is numerically difficult or not feasible.

- When there is a natural division of the parameters. That is the case for likelihood functions, where the parameter space naturally divides into parameter subsets, e.g. corresponding to linear effects, variances and covariances with different influence on the likelihood value.

- To improve optimization time in some cases, see [@hu:2002] for an example.

- Compared to joint optimization, alternating optimization may be better in bypassing local optima, see [@bezdek:2002].

### What are the properties of alternating optimization?

Alternating optimization under certain conditions on $f$ can convergence to the global optimum. However, the set of possible solutions also contains saddle points [@bezdek:1987]. @bezdek:2003 provides additional details on the convergence rate.

## Application

As an application, we consider minimizing the [Himmelblau's function](https://en.wikipedia.org/wiki/Himmelblau%27s_function) in $n = 2$ dimensions with parameter constraints $-5 \leq x_1, x_2 \leq 5$:

```{r, ao demo}
library("ao")
himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
ao(
  f = himmelblau, p = c(0, 0), partition = list(1, 2),
  base_optimizer = optimizeR::Optimizer$new(
    which = "stats::optim", lower = -5, upper = 5, method = "L-BFGS-B"
  )
)
```

The call minimizes `f` by alternating optimizing with respect to each parameter separately, where the parameters all are initialized at the value 0. 

## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{merge_results}
\alias{merge_results}
\title{Merge optimization results}
\usage{
merge_results(
  results,
  minimize = TRUE,
  add_details = TRUE,
  processes = data.frame()
)
}
\arguments{
\item{results}{[\code{list}]\cr
A \code{list} of outputs from \code{\link[ao]{ao}}.}

\item{minimize}{[\code{logical(1)}]\cr
Minimize during the AO process?

If \code{FALSE}, maximization is performed.}

\item{add_details}{[\code{logical(1)}]\cr
Add details about the AO process to the output?}

\item{processes}{[\code{data.frame}]\cr
A \code{data.frame} describing how the different processes were specified.}
}
\value{
A \code{list}, see section "Output value" on the \code{\link[ao]{ao}} page.
}
\description{
This helper function merges the results of multiple AO processes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ao.R
\name{ao}
\alias{ao}
\title{Alternating Optimization}
\usage{
ao(
  f,
  initial,
  target = NULL,
  npar = NULL,
  gradient = NULL,
  hessian = NULL,
  ...,
  partition = "sequential",
  new_block_probability = 0.3,
  minimum_block_number = 1,
  minimize = TRUE,
  lower = NULL,
  upper = NULL,
  iteration_limit = Inf,
  seconds_limit = Inf,
  tolerance_value = 1e-06,
  tolerance_parameter = 1e-06,
  tolerance_parameter_norm = function(x, y) sqrt(sum((x - y)^2)),
  tolerance_history = 1,
  base_optimizer = optimizeR::Optimizer$new("stats::optim", method = "L-BFGS-B", control
    = list(maxit = 10)),
  verbose = FALSE,
  hide_warnings = TRUE,
  add_details = TRUE
)
}
\arguments{
\item{f}{[\code{function}]\cr
A \code{function} to be optimized, returning a single \code{numeric} value.

The first argument of \code{f} should be a \code{numeric} of the same length
as \code{initial}, optionally followed by any other arguments specified by
the \code{...} argument.

If \code{f} is optimized over an argument other than the first, or over more
than one argument, specify this with the \code{target} argument.}

\item{initial}{[\code{numeric()} | \code{list()}]\cr
The starting parameter values for the target argument(s).

This can also be a \code{list} of multiple starting parameter values, see details.}

\item{target}{[\code{character()} | \code{NULL}]\cr
The name(s) of the argument(s) over which \code{f} is optimized.

This can only be \code{numeric} arguments.

If \code{NULL} (default), the first argument of \code{f} is optimized.}

\item{npar}{[\code{integer()}]\cr
The length(s) of the target argument(s).

Must be specified if more than one target argument is specified via
the \code{target} argument.

Can be \code{NULL} if there is only one target argument, in which case \code{npar} is
set to \code{length(initial)}.}

\item{gradient}{[\code{function} | \code{NULL}]\cr
An optional \code{function} that returns the gradient of \code{f}.

The function signature of \code{gradient} must be identical to \code{f}.

Ignored if \code{base_optimizer} does not support custom gradient.}

\item{hessian}{[\code{function} | \code{NULL}]\cr
An optional \code{function} that returns the Hessian of \code{f}.

The function signature of \code{hessian} must be identical to \code{f}.

Ignored if \code{base_optimizer} does not support custom Hessian.}

\item{...}{Additional arguments to be passed to \code{f} (and \code{gradient}).}

\item{partition}{[\code{character(1)} | \code{list()}]\cr
Defines the parameter partition. It can be
\itemize{
\item \code{"sequential"} for treating each parameter separately,
\item \code{"random"} for a random partition in each iteration,
\item \code{"none"} for no partition (which is equivalent to joint optimization),
\item or a \code{list} of vectors of parameter indices, specifying a custom
partition for the AO process.
}

This can also be a \code{list} of multiple partition definitions, see details.}

\item{new_block_probability}{[\code{numeric(1)}]\cr
Only relevant if \code{partition = "random"}.

The probability of creating a new parameter block when creating a random
partition.

Values close to 0 result in larger parameter blocks, values close to 1
result in smaller parameter blocks.}

\item{minimum_block_number}{[\code{integer(1)}]\cr
Only relevant if \code{partition = "random"}.

The minimum number of blocks in random partitions.}

\item{minimize}{[\code{logical(1)}]\cr
Minimize during the AO process?

If \code{FALSE}, maximization is performed.}

\item{lower, upper}{[\code{numeric()} | \code{NULL}]\cr
Optional lower and upper parameter bounds.

Ignored if \code{base_optimizer} does not support parameter bounds.}

\item{iteration_limit}{[\code{integer(1)} | \code{Inf}]\cr
The maximum number of iterations through the parameter partition before
the AO process is terminated.

Can also be \code{Inf} for no iteration limit.}

\item{seconds_limit}{[\code{numeric(1)}]\cr
The time limit in seconds before the AO process is terminated.

Can also be \code{Inf} for no time limit.

Note that this stopping criterion is only checked \emph{after} a sub-problem is
solved and not \emph{within} solving a sub-problem, so the actual process
time can exceed this limit.}

\item{tolerance_value}{[\code{numeric(1)}]\cr
A non-negative tolerance value. The AO process terminates
if the absolute difference between the current function value and the value
from \code{tolerance_history} iterations earlier is smaller than
\code{tolerance_value}.

Can be \code{0} for no value threshold.}

\item{tolerance_parameter}{[\code{numeric(1)}]\cr
A non-negative tolerance value. The AO process terminates if
the distance between the current estimate and the estimate from
\code{tolerance_history} iterations earlier is smaller than
\code{tolerance_parameter}.

Can be \code{0} for no parameter threshold.

By default, the distance is measured using the Euclidean norm, but another
norm can be specified via the \code{tolerance_parameter_norm} argument.}

\item{tolerance_parameter_norm}{[\code{function}]\cr
The norm that measures the distance between two estimates. If the distance
is smaller than
\code{tolerance_parameter}, the AO process is terminated.

It must be of the form \code{function(x, y)} for two vector inputs
\code{x} and \code{y}, and return a single \code{numeric} value.
By default, the Euclidean norm \code{function(x, y) sqrt(sum((x - y)^2))}
is used.}

\item{tolerance_history}{[\code{integer(1)}]\cr
The number of iterations to look back to determine whether
\code{tolerance_value} or \code{tolerance_parameter} has been reached.}

\item{base_optimizer}{[\code{Optimizer} | \code{list()}]\cr
An \code{Optimizer} object, which can be created via
\code{\link[optimizeR]{Optimizer}}. It numerically solves the sub-problems.

By default, the \code{\link[stats]{optim}} optimizer with
\code{method = "L-BFGS-B"} is used. The default value for the number of
iterations within each block is 10.

This can also be a \code{list} of multiple base optimizers, see details.}

\item{verbose}{[\code{logical(1)}]\cr
Print tracing details during the AO process?

Not supported when using multiple processes, see details.}

\item{hide_warnings}{[\code{logical(1)}]\cr
Hide warnings during the AO process?}

\item{add_details}{[\code{logical(1)}]\cr
Add details about the AO process to the output?}
}
\value{
A \code{list} with the following elements:
\itemize{
\item \code{estimate} is the parameter vector at termination.
\item \code{estimate_split} is \code{estimate} split by \code{target} (only when applicable).
\item \code{value} is the function value at termination.
\item \code{details} is a \code{data.frame} with information about the AO process:
For each iteration (column \code{iteration}) it contains the function value
(column \code{value}), parameter values (columns starting with \code{p} followed by
the parameter index), the active parameter block (columns starting with \code{b}
followed by the parameter index, where \code{1} stands for a parameter contained
in the active parameter block and \code{0} if not), and computation times in
seconds (column \code{seconds}). Only available if \code{add_details = TRUE}.
\item \code{seconds} is the overall computation time in seconds.
\item \code{stopping_reason} is a message explaining why the AO process terminated.
}

In the case of multiple processes, the output changes slightly, see details.
}
\description{
Alternating optimization (AO) is an iterative process for optimizing a
real-valued function jointly over all its parameters by alternating
restricted optimization over parameter partitions.
}
\details{
\subsection{Multiple processes}{

AO can suffer from local optima. To increase the likelihood of finding a
better optimum, you can specify:
\itemize{
\item multiple starting parameters
\item multiple parameter partitions
\item multiple base optimizers
}

Use the \code{initial}, \code{partition}, and/or \code{base_optimizer} arguments to provide
a \code{list} of possible values for each parameter. Each combination of initial
values, parameter partitions, and base optimizers will create a separate AO
process.
\subsection{Output value}{

In the case of multiple processes, the output values refer to the best
AO process with respect to function value.

If \code{add_details = TRUE}, the following elements are added:
\itemize{
\item \code{estimates} is a \code{list} of optimal parameters in each process.
\item \code{values} is a \code{list} of optimal function values in each process.
\item \code{details} combines details of the single processes and has an
additional column \code{process} with an index for the different processes.
\item \code{seconds_each} gives the computation time in seconds for each process.
\item \code{stopping_reasons} gives the termination message for each process.
\item \code{processes} gives details on how the processes were specified.
}
}

\subsection{Parallel computation}{

By default, processes run sequentially. However, since they are independent,
they can be parallelized. To enable parallel computation, use the
\href{https://future.futureverse.org/}{\code{{future}} framework}. For example, run the
following \emph{before} the \code{ao()} call:
\preformatted{
future::plan(future::multisession, workers = 4)
}
}

\subsection{Progress updates}{

When using multiple processes, setting \code{verbose = TRUE} to print tracing
details during AO is not supported. However, you can still track the progress
using the \href{https://progressr.futureverse.org/}{\code{{progressr}} framework}.
For example, run the following \emph{before} the \code{ao()} call:
\preformatted{
progressr::handlers(global = TRUE)
progressr::handlers(
  progressr::handler_progress(":percent :eta :message")
)
}
}

}
}
\examples{
# Example 1: Minimization of Himmelblau's function --------------------------

himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
ao(f = himmelblau, initial = c(0, 0))

# Example 2: Maximization of 2-class Gaussian mixture log-likelihood --------

normal_mixture_loglik_uc = function(mu, logsd, eta, data) {
  sd <- exp(logsd)
  e <- exp(eta[1])
  den <- 1 + e
  q1 <- e / den
  q2 <- 1 / den
  l1 <- log(q1) + dnorm(data, mu[1], sd[1], log = TRUE)
  l2 <- log(q2) + dnorm(data, mu[2], sd[2], log = TRUE)
  m <- pmax(l1, l2)
  sum(m + log(exp(l1 - m) + exp(l2 - m)))
}

set.seed(123)

data <- datasets::faithful$eruptions

fit <- ao(
  f = normal_mixture_loglik_uc,
  initial = c(mean(data) + c(-1, 1), rep(log(sd(data)), 2), 0),
  target = c("mu", "logsd", "eta"),
  npar = c(2, 2, 1),
  data = data,
  partition = "random",
  base_optimizer = optimizeR::Optimizer$new("ucminf::ucminf"),
  minimize = FALSE,
  add_details = FALSE
)

(muhat <- fit$estimate_split$mu)
(sdhat <- exp(fit$estimate_split$logsd))
e <- exp(fit$estimate_split$eta)
den <- 1 + e
(qhat <- c(e / den, 1 / den))

# Example 3: Constrained Optimization in the Setting of Example 2 -----------

# target arguments:
# - class means mu (2, unrestricted)
# - class standard deviations sd (2, must be non-negative)
# - class proportion lambda (only 1 for identification, must be in [0, 1])

normal_mixture_loglik <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}

set.seed(123)

ao(
  f = normal_mixture_loglik,
  initial = runif(5),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = list("sequential", "random", "none"),
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1),
  add_details = FALSE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{generate_random_partition}
\alias{generate_random_partition}
\title{Generate random partition}
\usage{
generate_random_partition(x, p, min)
}
\arguments{
\item{x}{[\code{integer}]\cr
The parameter indices.}

\item{p}{[\code{numeric(1)}]\cr
The probability of generating a new block.}

\item{min}{[\code{integer(1)}]\cr
The minimum number of blocks.}
}
\value{
A \code{list}, a random partition of \code{x}.
}
\description{
This helper function generates a random parameter partition, which is used
for the randomized AO procedure.
}
\author{
Siddhartha Chib
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{split_by_target}
\alias{split_by_target}
\title{Split estimate by target}
\usage{
split_by_target(estimate, target = NULL, npar)
}
\arguments{
\item{estimate}{[\code{numeric}]\cr
A parameter vector.}

\item{target}{[\code{character()} | \code{NULL}]\cr
The name(s) of the argument(s) over which \code{f} is optimized.

This can only be \code{numeric} arguments.

If \code{NULL} (default), the first argument of \code{f} is optimized.}

\item{npar}{[\code{integer()}]\cr
The length(s) of the target argument(s).

Must be specified if more than one target argument is specified via
the \code{target} argument.

Can be \code{NULL} if there is only one target argument, in which case \code{npar} is
set to \code{length(initial)}.}
}
\value{
A (named) \code{list}, a partition of \code{estimate} according to \code{npar}.
}
\description{
This helper function splits the solution by target parameters (if provided),
which is used for the output.
}
\author{
Siddhartha Chib
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process.R
\name{Process}
\alias{Process}
\title{Process Object}
\description{
This object specifies an AO process.
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{npar}}{[\code{integer()}]\cr
The length(s) of the target argument(s).}

\item{\code{partition}}{[\code{character(1)} | \code{list()}]\cr
Defines the parameter partition. It can be
\itemize{
\item \code{"sequential"} for treating each parameter separately,
\item \code{"random"} for a random partition in each iteration,
\item \code{"none"} for no partition (which is equivalent to joint optimization),
\item or a \code{list} of vectors of parameter indices, specifying a custom
partition for the AO process.
}}

\item{\code{new_block_probability}}{[\code{numeric(1)}]\cr
Only relevant if \code{partition = "random"}.

The probability of creating a new parameter block when creating a random
partition.

Values close to 0 result in larger parameter blocks, values close to 1
result in smaller parameter blocks.}

\item{\code{minimum_block_number}}{[\code{integer(1)}]\cr
Only relevant if \code{partition = "random"}.

The minimum number of blocks in random partitions.}

\item{\code{verbose}}{[\code{logical(1)}]\cr
Print tracing details during the AO process?}

\item{\code{minimize}}{[\code{logical(1)}]\cr
Minimize during the AO process?

If \code{FALSE}, maximization is performed.}

\item{\code{iteration_limit}}{[\code{integer(1)} | \code{Inf}]\cr
The maximum number of iterations through the parameter partition before
the AO process is terminated.

Can also be \code{Inf} for no iteration limit.}

\item{\code{seconds_limit}}{[\code{numeric(1)}]\cr
The time limit in seconds before the AO process is terminated.

Can also be \code{Inf} for no time limit.

Note that this stopping criterion is only checked \emph{after} a sub-problem is
solved and not \emph{within} solving a sub-problem, so the actual process
time can exceed this limit.}

\item{\code{tolerance_value}}{[\code{numeric(1)}]\cr
A non-negative tolerance value. The AO process terminates
if the absolute difference between the current function value and the
value from \code{tolerance_history} iterations earlier is smaller than
\code{tolerance_value}.

Can be \code{0} for no value threshold.}

\item{\code{tolerance_parameter}}{[\code{numeric(1)}]\cr
A non-negative tolerance value. The AO process terminates if
the distance between the current estimate and the estimate from
\code{tolerance_history} iterations earlier is smaller than
\code{tolerance_parameter}.

Can be \code{0} for no parameter threshold.

By default, the distance is measured using the Euclidean norm, but another
norm can be specified via the \code{tolerance_parameter_norm} field.}

\item{\code{tolerance_parameter_norm}}{[\code{function}]\cr
The norm that measures the distance between two estimates. If the distance
is smaller than
\code{tolerance_parameter}, the AO process is terminated.

It must be of the form \code{function(x, y)} for two vector inputs
\code{x} and \code{y}, and return a single \code{numeric} value.
By default, the Euclidean norm \code{function(x, y) sqrt(sum((x - y)^2))}
is used.}

\item{\code{tolerance_history}}{[\code{integer(1)}]\cr
The number of iterations to look back to determine whether
\code{tolerance_value} or \code{tolerance_parameter} has been reached.}

\item{\code{add_details}}{[\code{logical(1)}]\cr
Add details about the AO process to the output?}

\item{\code{iteration}}{[\code{integer(1)}]\cr
The current iteration number.}

\item{\code{block}}{[\code{integer()}]\cr
The currently active parameter block, represented as parameter indices.}

\item{\code{output}}{[\code{list()}, read-only]\cr
The output of the AO process, which is a \code{list} with the following
elements:
\itemize{
\item \code{estimate} is the parameter vector at termination.
\item \code{value} is the function value at termination.
\item \code{details} is a \code{data.frame} with full information about the
AO process.
For each iteration (column \code{iteration}) it contains the function value
(column \code{value}), parameter values (columns starting with \code{p}
followed by the parameter index), the active parameter block
(columns starting with \code{b} followed by the parameter index, where
\code{1} stands for a parameter contained in the active parameter block
and \code{0} if not), and computation times in seconds (column \code{seconds}).
Only available if \code{add_details = TRUE}.
\item \code{seconds} is the overall computation time in seconds.
\item \code{stopping_reason} is a message explaining why the AO process
terminated.
}}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Process-new}{\code{Process$new()}}
\item \href{#method-Process-print_status}{\code{Process$print_status()}}
\item \href{#method-Process-initialize_details}{\code{Process$initialize_details()}}
\item \href{#method-Process-update_details}{\code{Process$update_details()}}
\item \href{#method-Process-get_partition}{\code{Process$get_partition()}}
\item \href{#method-Process-get_details}{\code{Process$get_details()}}
\item \href{#method-Process-get_value}{\code{Process$get_value()}}
\item \href{#method-Process-get_value_latest}{\code{Process$get_value_latest()}}
\item \href{#method-Process-get_value_best}{\code{Process$get_value_best()}}
\item \href{#method-Process-get_parameter}{\code{Process$get_parameter()}}
\item \href{#method-Process-get_parameter_latest}{\code{Process$get_parameter_latest()}}
\item \href{#method-Process-get_parameter_best}{\code{Process$get_parameter_best()}}
\item \href{#method-Process-get_seconds}{\code{Process$get_seconds()}}
\item \href{#method-Process-get_seconds_total}{\code{Process$get_seconds_total()}}
\item \href{#method-Process-check_stopping}{\code{Process$check_stopping()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-new"></a>}}
\if{latex}{\out{\hypertarget{method-Process-new}{}}}
\subsection{Method \code{new()}}{
Creates a new object of this \link[R6:R6Class]{R6} class.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$new(
  target = NULL,
  npar = integer(),
  partition = "sequential",
  new_block_probability = 0.3,
  minimum_block_number = 1,
  verbose = FALSE,
  minimize = TRUE,
  iteration_limit = Inf,
  seconds_limit = Inf,
  tolerance_value = 1e-06,
  tolerance_parameter = 1e-06,
  tolerance_parameter_norm = function(x, y) sqrt(sum((x - y)^2)),
  tolerance_history = 1,
  add_details = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{target}}{[\code{character()} | \code{NULL}]\cr
The name(s) of the argument(s) over which \code{f} is optimized.

This can only be \code{numeric} arguments.

If \code{NULL} (default), the first argument of \code{f} is optimized.}

\item{\code{npar}}{[\code{integer()}]\cr
The length(s) of the target argument(s).}

\item{\code{partition}}{[\code{character(1)} | \code{list()}]\cr
Defines the parameter partition. It can be
\itemize{
\item \code{"sequential"} for treating each parameter separately,
\item \code{"random"} for a random partition in each iteration,
\item \code{"none"} for no partition (which is equivalent to joint optimization),
\item or a \code{list} of vectors of parameter indices, specifying a custom
partition for the AO process.
}}

\item{\code{new_block_probability}}{[\code{numeric(1)}]\cr
Only relevant if \code{partition = "random"}.

The probability of creating a new parameter block when creating a random
partition.

Values close to 0 result in larger parameter blocks, values close to 1
result in smaller parameter blocks.}

\item{\code{minimum_block_number}}{[\code{integer(1)}]\cr
Only relevant if \code{partition = "random"}.

The minimum number of blocks in random partitions.}

\item{\code{verbose}}{[\code{logical(1)}]\cr
Print tracing details during the AO process?}

\item{\code{minimize}}{[\code{logical(1)}]\cr
Minimize during the AO process?

If \code{FALSE}, maximization is performed.}

\item{\code{iteration_limit}}{[\code{integer(1)} | \code{Inf}]\cr
The maximum number of iterations through the parameter partition before
the AO process is terminated.

Can also be \code{Inf} for no iteration limit.}

\item{\code{seconds_limit}}{[\code{numeric(1)}]\cr
The time limit in seconds before the AO process is terminated.

Can also be \code{Inf} for no time limit.

Note that this stopping criterion is only checked \emph{after} a sub-problem is
solved and not \emph{within} solving a sub-problem, so the actual process
time can exceed this limit.}

\item{\code{tolerance_value}}{[\code{numeric(1)}]\cr
A non-negative tolerance value. The AO process terminates
if the absolute difference between the current function value and the value
from \code{tolerance_history} iterations earlier is smaller than
\code{tolerance_value}.

Can be \code{0} for no value threshold.}

\item{\code{tolerance_parameter}}{[\code{numeric(1)}]\cr
A non-negative tolerance value. The AO process terminates if
the distance between the current estimate and the estimate from
\code{tolerance_history} iterations earlier is smaller than
\code{tolerance_parameter}.

Can be \code{0} for no parameter threshold.

By default, the distance is measured using the Euclidean norm, but another
norm can be specified via the \code{tolerance_parameter_norm} field.}

\item{\code{tolerance_parameter_norm}}{[\code{function}]\cr
The norm that measures the distance between two estimates. If the distance
is smaller than
\code{tolerance_parameter}, the AO process is terminated.

It must be of the form \code{function(x, y)} for two vector inputs
\code{x} and \code{y}, and return a single \code{numeric} value.
By default, the Euclidean norm \code{function(x, y) sqrt(sum((x - y)^2))}
is used.}

\item{\code{tolerance_history}}{[\code{integer(1)}]\cr
The number of iterations to look back to determine whether
\code{tolerance_value} or \code{tolerance_parameter} has been reached.}

\item{\code{add_details}}{[\code{logical(1)}]\cr
Add details about the AO process to the output?}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-print_status"></a>}}
\if{latex}{\out{\hypertarget{method-Process-print_status}{}}}
\subsection{Method \code{print_status()}}{
Prints a status message.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$print_status(message, message_type = 8, verbose = self$verbose)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{message}}{[\code{character(1)}]\cr
A status message.}

\item{\code{message_type}}{[\code{integer(1)}]\cr
The message type, one of the following:
\itemize{
\item \code{1} for \code{cli::cli_h1()}
\item \code{2} for \code{cli::cli_h2()}
\item \code{3} for \code{cli::cli_h3()}
\item \code{4} for \code{cli::cli_alert_success()}
\item \code{5} for \code{cli::cli_alert_info()}
\item \code{6} for \code{cli::cli_alert_warning()}
\item \code{7} for \code{cli::cli_alert_danger()}
\item \code{8} for \code{cli::cat_line()}
}}

\item{\code{verbose}}{[\code{logical(1)}]\cr
Print tracing details during the AO process?}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-initialize_details"></a>}}
\if{latex}{\out{\hypertarget{method-Process-initialize_details}{}}}
\subsection{Method \code{initialize_details()}}{
Initializes the \code{details} part of the output.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$initialize_details(initial_parameter, initial_value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{initial_parameter}}{[\code{numeric()}]\cr
The starting parameter values for the AO process.}

\item{\code{initial_value}}{[\code{numeric(1)}]\cr
The function value at the initial parameters.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-update_details"></a>}}
\if{latex}{\out{\hypertarget{method-Process-update_details}{}}}
\subsection{Method \code{update_details()}}{
Updates the \code{details} part of the output.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$update_details(
  value,
  parameter_block,
  seconds,
  error,
  error_message,
  block = self$block
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{[\code{numeric(1)}]\cr
The updated function value.}

\item{\code{parameter_block}}{[\code{numeric()}]\cr
The updated parameter values for the active parameter block.}

\item{\code{seconds}}{[\code{numeric(1)}]\cr
The time in seconds for solving the sub-problem.}

\item{\code{error}}{[\code{logical(1)}]\cr
Did solving the sub-problem result in an error?}

\item{\code{error_message}}{[\code{character(1)}]\cr
An error message if \code{error = TRUE}.}

\item{\code{block}}{[\code{integer()}]\cr
The currently active parameter block, represented as parameter indices.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_partition"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_partition}{}}}
\subsection{Method \code{get_partition()}}{
Get a parameter partition.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_partition()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_details"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_details}{}}}
\subsection{Method \code{get_details()}}{
Get the \code{details} part of the output.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_details(
  which_iteration = NULL,
  which_block = NULL,
  which_column = c("iteration", "value", "parameter", "block", "seconds")
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{which_iteration}}{[\code{integer()}]\cr
Selects the iteration(s).

Can also be \code{NULL} to select all iterations.}

\item{\code{which_block}}{[\code{character(1)} | \code{integer()}]\cr
Selects the parameter block in the partition and can be one of
\itemize{
\item \code{"first"} for the first parameter block,
\item \code{"last"} for the last parameter block,
\item an \code{integer} vector of parameter indices,
\item or \code{NULL} for all parameter blocks.
}}

\item{\code{which_column}}{[\code{character()}]\cr
Selects the columns in the \code{details} part of the output and can be one or
more of
\itemize{
\item \code{"iteration"},
\item \code{"value"},
\item \code{"parameter"},
\item \code{"block"},
\item and \code{"seconds"}.
}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_value"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_value}{}}}
\subsection{Method \code{get_value()}}{
Get the function value in different steps of the AO process.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_value(
  which_iteration = NULL,
  which_block = NULL,
  keep_iteration_column = FALSE,
  keep_block_columns = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{which_iteration}}{[\code{integer()}]\cr
Selects the iteration(s).

Can also be \code{NULL} to select all iterations.}

\item{\code{which_block}}{[\code{character(1)} | \code{integer()}]\cr
Selects the parameter block in the partition and can be one of
\itemize{
\item \code{"first"} for the first parameter block,
\item \code{"last"} for the last parameter block,
\item an \code{integer} vector of parameter indices,
\item or \code{NULL} for all parameter blocks.
}}

\item{\code{keep_iteration_column}}{[\code{logical(1)}]\cr
Keep the column containing the information about the iteration in the output?}

\item{\code{keep_block_columns}}{[\code{logical(1)}]\cr
Keep the column containing the information about the active parameter block
in the output?}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_value_latest"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_value_latest}{}}}
\subsection{Method \code{get_value_latest()}}{
Get the function value in the latest step of the AO process.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_value_latest()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_value_best"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_value_best}{}}}
\subsection{Method \code{get_value_best()}}{
Get the optimum function value in the AO process.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_value_best()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_parameter"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_parameter}{}}}
\subsection{Method \code{get_parameter()}}{
Get the parameter values in different steps of the AO process.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_parameter(
  which_iteration = self$iteration,
  which_block = NULL,
  keep_iteration_column = FALSE,
  keep_block_columns = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{which_iteration}}{[\code{integer()}]\cr
Selects the iteration(s).

Can also be \code{NULL} to select all iterations.}

\item{\code{which_block}}{[\code{character(1)} | \code{integer()}]\cr
Selects the parameter block in the partition and can be one of
\itemize{
\item \code{"first"} for the first parameter block,
\item \code{"last"} for the last parameter block,
\item an \code{integer} vector of parameter indices,
\item or \code{NULL} for all parameter blocks.
}}

\item{\code{keep_iteration_column}}{[\code{logical(1)}]\cr
Keep the column containing the information about the iteration in the output?}

\item{\code{keep_block_columns}}{[\code{logical(1)}]\cr
Keep the column containing the information about the active parameter block
in the output?}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_parameter_latest"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_parameter_latest}{}}}
\subsection{Method \code{get_parameter_latest()}}{
Get the parameter value in the latest step of the AO process.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_parameter_latest(parameter_type = "full")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{parameter_type}}{[\code{character(1)}]\cr
Selects the parameter type and can be one of
\itemize{
\item \code{"full"} (default) to get the full parameter vector,
\item \code{"block"} to get the parameter values for the current block,
i.e., the parameters with the indices \code{self$block}
\item \code{"fixed"} to get the parameter values which are currently fixed,
i.e., all except for those with the indices \code{self$block}
}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_parameter_best"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_parameter_best}{}}}
\subsection{Method \code{get_parameter_best()}}{
Get the optimum parameter value in the AO process.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_parameter_best(parameter_type = "full")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{parameter_type}}{[\code{character(1)}]\cr
Selects the parameter type and can be one of
\itemize{
\item \code{"full"} (default) to get the full parameter vector,
\item \code{"block"} to get the parameter values for the current block,
i.e., the parameters with the indices \code{self$block}
\item \code{"fixed"} to get the parameter values which are currently fixed,
i.e., all except for those with the indices \code{self$block}
}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_seconds"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_seconds}{}}}
\subsection{Method \code{get_seconds()}}{
Get the optimization time in seconds in different steps of the AO process.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_seconds(
  which_iteration = NULL,
  which_block = NULL,
  keep_iteration_column = FALSE,
  keep_block_columns = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{which_iteration}}{[\code{integer()}]\cr
Selects the iteration(s).

Can also be \code{NULL} to select all iterations.}

\item{\code{which_block}}{[\code{character(1)} | \code{integer()}]\cr
Selects the parameter block in the partition and can be one of
\itemize{
\item \code{"first"} for the first parameter block,
\item \code{"last"} for the last parameter block,
\item an \code{integer} vector of parameter indices,
\item or \code{NULL} for all parameter blocks.
}}

\item{\code{keep_iteration_column}}{[\code{logical(1)}]\cr
Keep the column containing the information about the iteration in the output?}

\item{\code{keep_block_columns}}{[\code{logical(1)}]\cr
Keep the column containing the information about the active parameter block
in the output?}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-get_seconds_total"></a>}}
\if{latex}{\out{\hypertarget{method-Process-get_seconds_total}{}}}
\subsection{Method \code{get_seconds_total()}}{
Get the total optimization time in seconds of the AO process.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$get_seconds_total()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Process-check_stopping"></a>}}
\if{latex}{\out{\hypertarget{method-Process-check_stopping}{}}}
\subsection{Method \code{check_stopping()}}{
Checks if the AO process can be terminated.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Process$check_stopping()}\if{html}{\out{</div>}}
}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ao-package.R
\docType{package}
\name{ao-package}
\alias{ao-package}
\title{ao: Alternating Optimization}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Implementation of an iterative process that optimizes a function by alternately performing restricted optimization over parameter subsets. Instead of solving one joint optimization problem, alternating optimization breaks it into smaller sub-problems. This approach can make optimization feasible when joint optimization is too difficult.
}
\examples{
# Example 1: Minimization of Himmelblau's function --------------------------

himmelblau <- function(x) (x[1]^2 + x[2] - 11)^2 + (x[1] + x[2]^2 - 7)^2
ao(f = himmelblau, initial = c(0, 0))

# Example 2: Maximization of 2-class Gaussian mixture log-likelihood --------

normal_mixture_loglik_uc = function(mu, logsd, eta, data) {
  sd <- exp(logsd)
  e <- exp(eta[1])
  den <- 1 + e
  q1 <- e / den
  q2 <- 1 / den
  l1 <- log(q1) + dnorm(data, mu[1], sd[1], log = TRUE)
  l2 <- log(q2) + dnorm(data, mu[2], sd[2], log = TRUE)
  m <- pmax(l1, l2)
  sum(m + log(exp(l1 - m) + exp(l2 - m)))
}

set.seed(123)

data <- datasets::faithful$eruptions

fit <- ao(
  f = normal_mixture_loglik_uc,
  initial = c(mean(data) + c(-1, 1), rep(log(sd(data)), 2), 0),
  target = c("mu", "logsd", "eta"),
  npar = c(2, 2, 1),
  data = data,
  partition = "random",
  base_optimizer = optimizeR::Optimizer$new("ucminf::ucminf"),
  minimize = FALSE,
  add_details = FALSE
)

(muhat <- fit$estimate_split$mu)
(sdhat <- exp(fit$estimate_split$logsd))
e <- exp(fit$estimate_split$eta)
den <- 1 + e
(qhat <- c(e / den, 1 / den))

# Example 3: Constrained Optimization in the Setting of Example 2 -----------

# target arguments:
# - class means mu (2, unrestricted)
# - class standard deviations sd (2, must be non-negative)
# - class proportion lambda (only 1 for identification, must be in [0, 1])

normal_mixture_loglik <- function(mu, sd, lambda, data) {
  c1 <- lambda * dnorm(data, mu[1], sd[1])
  c2 <- (1 - lambda) * dnorm(data, mu[2], sd[2])
  sum(log(c1 + c2))
}

set.seed(123)

ao(
  f = normal_mixture_loglik,
  initial = runif(5),
  target = c("mu", "sd", "lambda"),
  npar = c(2, 2, 1),
  data = datasets::faithful$eruptions,
  partition = list("sequential", "random", "none"),
  minimize = FALSE,
  lower = c(-Inf, -Inf, 0, 0, 0),
  upper = c(Inf, Inf, Inf, Inf, 1),
  add_details = FALSE
)

}
\seealso{
Useful links:
\itemize{
  \item \url{https://loelschlaeger.de/ao/}
  \item \url{https://github.com/loelschlaeger/ao/}
  \item Report bugs at \url{https://github.com/loelschlaeger/ao/issues}
}

}
\author{
\strong{Maintainer}: Lennart Oelschläger \email{oelschlaeger.lennart@gmail.com} (\href{https://orcid.org/0000-0001-5421-9313}{ORCID})

Other contributors:
\itemize{
  \item Siddhartha Chib \email{chib@wustl.edu} [contributor]
}

}
\keyword{internal}
